"""Main module for running the drug response prediction experiment."""

import importlib
import json
import os
import shutil
import tempfile
import warnings
from typing import Any

import numpy as np
import pandas as pd
import torch
from sklearn.base import TransformerMixin

from .datasets.dataset import DrugResponseDataset, FeatureDataset
from .evaluation import evaluate, get_mode
from .models import MODEL_FACTORY, MULTI_DRUG_MODEL_FACTORY, SINGLE_DRUG_MODEL_FACTORY
from .models.drp_model import DRPModel
from .pipeline_function import pipeline_function

if importlib.util.find_spec("ray"):
    import ray
else:
    ray = None


def drug_response_experiment(
    models: list[type[DRPModel]],
    response_data: DrugResponseDataset,
    baselines: list[type[DRPModel]] | None = None,
    response_transformation: TransformerMixin | None = None,
    run_id: str = "",
    test_mode: str = "LPO",
    metric: str = "RMSE",
    n_cv_splits: int = 5,
    multiprocessing: bool = False,
    randomization_mode: list[str] | None = None,
    randomization_type: str = "permutation",
    cross_study_datasets: list[DrugResponseDataset] | None = None,
    n_trials_robustness: int = 0,
    path_out: str = "results/",
    overwrite: bool = False,
    path_data: str = "data",
    model_checkpoint_dir: str = "TEMPORARY",
) -> None:
    """
    Run the drug response prediction experiment. Save results to disc.

    :param models: list of model classes to compare
    :param baselines: list of baseline models. No randomization or robustness tests are run for the baseline models.
    :param response_data: drug response dataset
    :param response_transformation: normalizer to use for the response data
    :param metric: metric to use for hyperparameter optimization
    :param n_cv_splits: number of cross-validation splits
    :param multiprocessing: whether to use multiprocessing. This requires Ray to be installed.
    :param randomization_mode: list of randomization modes to do. Modes: SVCC, SVRC, SVCD, SVRD Can be a list of
        randomization tests e.g. 'SVCC SVCD'. Default is None, which means no randomization tests are run.

        * SVCC: Single View Constant for Cell Lines: in this mode, one experiment is done for every cell line view
            the model uses (e.g. gene expression, mutation, ...). For each experiment one cell line view is held
            constant while the others are randomized.
        * SVRC Single View Random for Cell Lines: in this mode, one experiment is done for every cell line view the
            model uses (e.g. gene expression, mutation, ...). For each experiment one cell line view is randomized while
            the others are held constant.
        * SVCD: Single View Constant for Drugs: in this mode, one experiment is done for every drug view the model
            uses (e.g. fingerprints, target_information, ...). For each experiment one drug view is held constant
            while the others are randomized.
        * SVRD: Single View Random for Drugs: in this mode, one experiment is done for every drug view the model uses
            (e.g. gene expression, target_information, ...). For each experiment one drug view is randomized while
            the others are held constant.

    :param randomization_type: type of randomization to use. Choose from "permutation" and "invariant".
        Default is "permutation".

        * "permutation": permute the features over the instances, keeping the distribution of the features the same
            but dissolving the relationship to the target
        * "invariant": the features are permuted in a way that a key characteristic of the feature is kept. In case of
            matrices, this is the mean and standard deviation of the feature view for this instance, for networks it
            is the degree distribution.

    :param cross_study_datasets: list of datasets for the cross-study prediction. The trained model is assessed for
        its generalization to these datasets. Default is None, which means no cross-study prediction is run.
    :param n_trials_robustness: number of trials to run for the robustness test. The robustness test is a test where
        models are retrained multiple times with varying seeds. Default is 0, which means no robustness test is run.
    :param path_out: path to the output directory
    :param run_id: identifier to save the results
    :param test_mode: test mode one of "LPO", "LCO", "LDO" (leave-pair-out, leave-cell-line-out, leave-drug-out)
    :param overwrite: whether to overwrite existing results
    :param path_data: path to the data directory, usually data/
    :param model_checkpoint_dir: directory to save model checkpoints. If "TEMPORARY", a temporary directory is created.
    :raises ValueError: if no cv splits are found
    """
    if baselines is None:
        baselines = []
    cross_study_datasets = cross_study_datasets or []
    result_path = os.path.join(path_out, run_id, test_mode)
    split_path = os.path.join(result_path, "splits")
    result_folder_exists = os.path.exists(result_path)
    if result_folder_exists and overwrite:
        # if results exists, delete them if overwrite is True
        print(f"Overwriting existing results at {result_path}")
        shutil.rmtree(result_path)

    if result_folder_exists and os.path.exists(split_path):
        # if the results exist and overwrite is false, load the cv splits.
        # The models will be trained on the existing cv splits.
        print(f"Loading existing cv splits from {split_path}")
        response_data.load_splits(path=split_path)
    else:
        # if the results do not exist, create the cv splits
        print(f"Creating cv splits at {split_path}")

        os.makedirs(result_path, exist_ok=True)

        response_data.remove_nan_responses()
        # if this line changes, also change it in pipeline: cv_split.py
        response_data.split_dataset(
            n_cv_splits=n_cv_splits,
            mode=test_mode,
            split_validation=True,
            validation_ratio=0.1,
            random_state=42,
        )
        response_data.save_splits(path=split_path)

    model_list = make_model_list(models + baselines, response_data)
    for model_name in model_list.keys():
        print(f"Running {model_name}")
        model_name, drug_id = get_model_name_and_drug_id(model_name)

        model_class = MODEL_FACTORY[model_name]
        if model_class in baselines:
            print("- Only Baseline Tests -")
            is_baseline = True
        else:
            print("- Full Test -")
            is_baseline = False

        predictions_path = generate_data_saving_path(
            model_name=model_name,
            drug_id=drug_id,
            result_path=result_path,
            suffix="predictions",
        )
        hpam_path = generate_data_saving_path(
            model_name=model_name,
            drug_id=drug_id,
            result_path=result_path,
            suffix="best_hpams",
        )
        parent_dir = os.path.dirname(predictions_path)

        model_hpam_set = model_class.get_hyperparameter_set()

        if response_data.cv_splits is None:
            raise ValueError("No cv splits found.")

        for split_index, split in enumerate(response_data.cv_splits):
            print(f"################# FOLD {split_index + 1}/{len(response_data.cv_splits)} " f"#################")

            prediction_file = os.path.join(predictions_path, f"predictions_split_{split_index}.csv")

            hpam_filename = f"best_hpams_split_{split_index}.json"
            hpam_save_path = os.path.join(hpam_path, hpam_filename)

            (
                train_dataset,
                validation_dataset,
                early_stopping_dataset,
                test_dataset,
            ) = get_datasets_from_cv_split(split, model_class, model_name, drug_id)

            model = model_class()

            if not os.path.isfile(
                prediction_file
            ):  # if this split has not been run yet (or for a single drug model, this drug_id)
                tuning_inputs = {
                    "model": model,
                    "train_dataset": train_dataset,
                    "validation_dataset": validation_dataset,
                    "early_stopping_dataset": early_stopping_dataset,
                    "hpam_set": model_hpam_set,
                    "response_transformation": response_transformation,
                    "metric": metric,
                    "path_data": path_data,
                    "model_checkpoint_dir": model_checkpoint_dir,
                }

                if multiprocessing:
                    tuning_inputs["ray_path"] = os.path.abspath(os.path.join(result_path, "raytune"))
                    best_hpams = hpam_tune_raytune(**tuning_inputs)
                else:
                    best_hpams = hpam_tune(**tuning_inputs)

                print(f"Best hyperparameters: {best_hpams}")
                print("Training model on full train and validation set to predict test set")
                # save best hyperparameters as json
                with open(
                    hpam_save_path,
                    "w",
                    encoding="utf-8",
                ) as f:
                    json.dump(best_hpams, f)

                train_dataset.add_rows(validation_dataset)  # use full train val set data for final training
                train_dataset.shuffle(random_state=42)

                test_dataset = train_and_predict(
                    model=model,
                    hpams=best_hpams,
                    path_data=path_data,
                    train_dataset=train_dataset,
                    prediction_dataset=test_dataset,
                    early_stopping_dataset=(early_stopping_dataset if model.early_stopping else None),
                    response_transformation=response_transformation,
                    model_checkpoint_dir=model_checkpoint_dir,
                )

                for cross_study_dataset in cross_study_datasets:
                    print(f"Cross study prediction on {cross_study_dataset.dataset_name}")
                    cross_study_dataset.remove_nan_responses()
                    cross_study_prediction(
                        dataset=cross_study_dataset,
                        model=model,
                        test_mode=test_mode,
                        train_dataset=train_dataset,
                        path_data=path_data,
                        early_stopping_dataset=(early_stopping_dataset if model.early_stopping else None),
                        response_transformation=response_transformation,
                        path_out=parent_dir,
                        split_index=split_index,
                        single_drug_id=(drug_id if model_name in SINGLE_DRUG_MODEL_FACTORY else None),
                    )

                test_dataset.to_csv(prediction_file)
            else:
                print(f"Split {split_index} already exists. Skipping.")
                with open(
                    hpam_save_path,
                    encoding="utf-8",
                ) as f:
                    best_hpams = json.load(f)
            if not is_baseline:
                if randomization_mode is not None:
                    print(f"Randomization tests for {model_class.get_model_name()}")
                    # if this line changes, it also needs to be changed in pipeline:
                    # randomization_split.py
                    randomization_test_views = get_randomization_test_views(
                        model=model, randomization_mode=randomization_mode
                    )
                    randomization_test(
                        randomization_test_views=randomization_test_views,
                        model=model,
                        hpam_set=best_hpams,
                        path_data=path_data,
                        train_dataset=train_dataset,
                        test_dataset=test_dataset,
                        early_stopping_dataset=(early_stopping_dataset if model.early_stopping else None),
                        path_out=parent_dir,
                        split_index=split_index,
                        randomization_type=randomization_type,
                        response_transformation=response_transformation,
                        model_checkpoint_dir=model_checkpoint_dir,
                    )
                if n_trials_robustness > 0:
                    print(f"Robustness test for {model_class.get_model_name()}")
                    robustness_test(
                        n_trials=n_trials_robustness,
                        model=model,
                        hpam_set=best_hpams,
                        path_data=path_data,
                        train_dataset=train_dataset,
                        test_dataset=test_dataset,
                        early_stopping_dataset=(early_stopping_dataset if model.early_stopping else None),
                        path_out=parent_dir,
                        split_index=split_index,
                        response_transformation=response_transformation,
                    )
    consolidate_single_drug_model_predictions(
        models=models,
        n_cv_splits=n_cv_splits,
        results_path=result_path,
        cross_study_datasets=cross_study_datasets,
        randomization_mode=randomization_mode,
        n_trials_robustness=n_trials_robustness,
        out_path=result_path,
    )
    print("Done!")


@pipeline_function
def consolidate_single_drug_model_predictions(
    models: list[type[DRPModel]],
    n_cv_splits: int,
    results_path: str,
    cross_study_datasets: list[DrugResponseDataset],
    randomization_mode: list[str] | None = None,
    n_trials_robustness: int = 0,
    out_path: str = "",
) -> None:
    """
    Consolidate single drug model predictions into a single file.

    :param models: list of model classes to compare, e.g., [SimpleNeuralNetwork, RandomForest]
    :param n_cv_splits: number of cross-validation splits, e.g., 5
    :param results_path: path to the results directory, e.g., results/
    :param cross_study_datasets: list of cross-study datasets, e.g., [CCLE, GDSC1]
    :param randomization_mode: list of randomization modes, e.g., ["SVCC", "SVRC"]
    :param n_trials_robustness: number of robustness trials, e.g., 10
    :param out_path: for the package, this is the same as results_path. For the pipeline, this is empty because it
        will be stored in the work directory.
    """
    for model in models:
        if model.get_model_name() in SINGLE_DRUG_MODEL_FACTORY:
            model_instance = MODEL_FACTORY[model.get_model_name()]()
            model_path = os.path.join(results_path, model.get_model_name())
            out_path = os.path.join(out_path, model.get_model_name())
            os.makedirs(os.path.join(out_path, "predictions"), exist_ok=True)
            if cross_study_datasets:
                os.makedirs(os.path.join(out_path, "cross_study"), exist_ok=True)
            if randomization_mode:
                os.makedirs(os.path.join(out_path, "randomization"), exist_ok=True)
            if n_trials_robustness:
                os.makedirs(os.path.join(out_path, "robustness"), exist_ok=True)

            for split in range(n_cv_splits):
                # Collect predictions for drugs across all scenarios (main, cross_study, robustness, randomization)
                predictions: Any = {
                    "main": [],
                    "cross_study": {},
                    "robustness": {},
                    "randomization": {},
                }
                # list all dirs in model_path/drugs
                drugs = [
                    d
                    for d in os.listdir(os.path.join(model_path, "drugs"))
                    if os.path.isdir(os.path.join(model_path, "drugs", d))
                ]
                for drug in drugs:
                    single_drug_prediction_path = os.path.join(model_path, "drugs", drug)

                    # Main predictions
                    predictions["main"].append(
                        pd.read_csv(
                            os.path.join(
                                single_drug_prediction_path,
                                "predictions",
                                f"predictions_split_{split}.csv",
                            ),
                            index_col=0,
                        )
                    )

                    # Cross study predictions
                    for cross_study_dataset in cross_study_datasets:
                        cross_study_prediction_path = os.path.join(single_drug_prediction_path, "cross_study")
                        f = f"cross_study_{cross_study_dataset.dataset_name}_split_{split}.csv"
                        if cross_study_dataset.dataset_name not in predictions["cross_study"]:
                            predictions["cross_study"][cross_study_dataset.dataset_name] = []
                        predictions["cross_study"][cross_study_dataset.dataset_name].append(
                            pd.read_csv(
                                os.path.join(cross_study_prediction_path, f),
                                index_col=0,
                            )
                        )

                    # Robustness predictions
                    for trial in range(n_trials_robustness):
                        robustness_path = os.path.join(single_drug_prediction_path, "robustness")
                        f = f"robustness_{trial + 1}_split_{split}.csv"
                        if trial not in predictions["robustness"]:
                            predictions["robustness"][trial] = []
                        predictions["robustness"][trial].append(
                            pd.read_csv(os.path.join(robustness_path, f), index_col=0)
                        )

                    # Randomization predictions
                    if randomization_mode is not None:
                        randomization_test_views = get_randomization_test_views(
                            model=model_instance,
                            randomization_mode=randomization_mode,
                        )
                        for view in randomization_test_views:
                            randomization_path = os.path.join(single_drug_prediction_path, "randomization")
                            f = f"randomization_{view}_split_{split}.csv"
                            if view not in predictions["randomization"]:
                                predictions["randomization"][view] = []
                            predictions["randomization"][view].append(
                                pd.read_csv(
                                    os.path.join(randomization_path, f),
                                    index_col=0,
                                )
                            )

                # Save the consolidated predictions
                pd.concat(predictions["main"], axis=0).to_csv(
                    os.path.join(
                        out_path,
                        "predictions",
                        f"predictions_split_{split}.csv",
                    )
                )

                for dataset_name, dataset_predictions in predictions["cross_study"].items():
                    pd.concat(dataset_predictions, axis=0).to_csv(
                        os.path.join(
                            out_path,
                            "cross_study",
                            f"cross_study_{dataset_name}_split_{split}.csv",
                        )
                    )

                for trial, trial_predictions in predictions["robustness"].items():
                    pd.concat(trial_predictions, axis=0).to_csv(
                        os.path.join(
                            out_path,
                            "robustness",
                            f"robustness_{trial + 1}_split_{split}.csv",
                        )
                    )

                for view, view_predictions in predictions["randomization"].items():
                    pd.concat(view_predictions, axis=0).to_csv(
                        os.path.join(
                            out_path,
                            "randomization",
                            f"randomization_{view}_split_{split}.csv",
                        )
                    )


def load_features(
    model: DRPModel, path_data: str, dataset: DrugResponseDataset
) -> tuple[FeatureDataset, FeatureDataset | None]:
    """
    Load and reduce cell line and drug features for a given dataset.

    :param model: model to use, e.g., SimpleNeuralNetwork
    :param path_data: path to the data directory, e.g., data/
    :param dataset: dataset to load features for, e.g., GDSC2
    :returns: tuple of cell line and, potentially, drug features
    """
    cl_features = model.load_cell_line_features(data_path=path_data, dataset_name=dataset.dataset_name)
    drug_features = model.load_drug_features(data_path=path_data, dataset_name=dataset.dataset_name)
    return cl_features, drug_features


@pipeline_function
def cross_study_prediction(
    dataset: DrugResponseDataset,
    model: DRPModel,
    test_mode: str,
    train_dataset: DrugResponseDataset,
    path_data: str,
    early_stopping_dataset: DrugResponseDataset | None,
    response_transformation: TransformerMixin | None,
    path_out: str,
    split_index: int,
    single_drug_id: str | None = None,
) -> None:
    """
    Run the drug response prediction experiment on a cross-study dataset to assess the generalizability of the model.

    :param dataset: cross-study dataset, e.g., GDSC1 if trained on GDSC2
    :param model: model to use, e.g, SimpleNeuralNetwork
    :param test_mode: test mode one of "LPO", "LCO", "LDO" (leave-pair-out, leave-cell-line-out,
        leave-drug-out)
    :param train_dataset: training dataset, e.g., GDSC2
    :param path_data: path to the data directory, e.g., data/
    :param early_stopping_dataset: early stopping dataset
    :param response_transformation: normalizer to use for the response data, e.g., StandardScaler
    :param path_out: path to the output directory, e.g., results/
    :param split_index: index of the split
    :param single_drug_id: drug id to use for single drug models None for global models
    :raises ValueError: if feature loading fails or if the test mode is invalid
    """
    dataset = dataset.copy()
    os.makedirs(os.path.join(path_out, "cross_study"), exist_ok=True)
    if response_transformation:
        dataset.transform(response_transformation)

    # load features
    try:
        cl_features, drug_features = load_features(model, path_data, dataset)
    except ValueError as e:
        warnings.warn(str(e), stacklevel=2)
        return

    cell_lines_to_keep = cl_features.identifiers if cl_features is not None else None

    drugs_to_keep: np.ndarray | None = None
    if single_drug_id is not None:
        drugs_to_keep = np.array([single_drug_id])
    elif drug_features is not None:
        drugs_to_keep = drug_features.identifiers

    print(
        f"Reducing cross study dataset ... feature data available for "
        f'{len(cell_lines_to_keep) if cell_lines_to_keep is not None else "all"} cell lines '
        f'and {len(drugs_to_keep)if drugs_to_keep is not None else "all"} drugs.'
    )

    # making sure there are no missing features. Only keep cell lines and drugs for which we have
    # a feature representation
    dataset.reduce_to(cell_line_ids=cell_lines_to_keep, drug_ids=drugs_to_keep)
    if early_stopping_dataset is not None:
        train_dataset.add_rows(early_stopping_dataset)
    # remove rows which overlap in the training. depends on the test mode
    if test_mode == "LPO":
        train_pairs = {
            f"{cl}_{drug}" for cl, drug in zip(train_dataset.cell_line_ids, train_dataset.drug_ids, strict=True)
        }
        dataset_pairs = [f"{cl}_{drug}" for cl, drug in zip(dataset.cell_line_ids, dataset.drug_ids, strict=True)]

        dataset.remove_rows(np.array([i for i, pair in enumerate(dataset_pairs) if pair in train_pairs]))
    elif test_mode == "LCO":
        train_cell_lines = train_dataset.cell_line_ids
        dataset.reduce_to(
            cell_line_ids=np.setdiff1d(dataset.cell_line_ids, train_cell_lines),
            drug_ids=None,
        )
    elif test_mode == "LDO":
        train_drugs = train_dataset.drug_ids
        dataset.reduce_to(
            cell_line_ids=None,
            drug_ids=np.setdiff1d(dataset.drug_ids, train_drugs),
        )
    else:
        raise ValueError(f"Invalid test mode: {test_mode}. Choose from LPO, LCO, LDO")
    if len(dataset) > 0:
        dataset.shuffle(random_state=42)
        dataset._predictions = model.predict(
            cell_line_ids=dataset.cell_line_ids,
            drug_ids=dataset.drug_ids,
            cell_line_input=cl_features,
            drug_input=drug_features,
        )
        if response_transformation:
            dataset._response = response_transformation.inverse_transform(dataset.response)
    else:
        dataset._predictions = np.array([])
    dataset.to_csv(
        os.path.join(
            path_out,
            "cross_study",
            f"cross_study_{dataset.dataset_name}_split_{split_index}.csv",
        )
    )


@pipeline_function
def get_randomization_test_views(model: DRPModel, randomization_mode: list[str]) -> dict[str, list[str]]:
    """
    Get the views to use for the randomization tests.

    * For SVCC, a single cell line view (e.g., gene expression) is held constant while the others are randomized.
    * For SVCD, a single drug view (e.g., fingerprints) is held constant while the others are randomized.
    * For SVRC, a single cell line view is randomized while the others are held constant.
    * For SVRD, a single drug view is randomized while the others are held constant.

    :param model: model to use, e.g., SimpleNeuralNetwork
    :param randomization_mode: list of randomization modes to do, e.g., ["SVCC", "SVRC"]
    :returns: dictionary of randomization test views
    """
    cell_line_views = model.cell_line_views
    drug_views = model.drug_views
    randomization_test_views = {}
    if "SVCC" in randomization_mode:
        for view in cell_line_views:
            randomization_test_views[f"SVCC_{view}"] = [v for v in cell_line_views if v != view]
    if "SVCD" in randomization_mode:
        for view in drug_views:
            randomization_test_views[f"SVCD_{view}"] = [v for v in drug_views if v != view]
    if "SVRC" in randomization_mode:
        for view in cell_line_views:
            randomization_test_views[f"SVRC_{view}"] = [view]
    if "SVRD" in randomization_mode:
        for view in drug_views:
            randomization_test_views[f"SVRD_{view}"] = [view]

    return randomization_test_views


def robustness_test(
    n_trials: int,
    model: DRPModel,
    hpam_set: dict,
    path_data: str,
    train_dataset: DrugResponseDataset,
    test_dataset: DrugResponseDataset,
    early_stopping_dataset: DrugResponseDataset | None,
    path_out: str,
    split_index: int,
    response_transformation: TransformerMixin | None = None,
    model_checkpoint_dir: str = "TEMPORARY",
):
    """
    Run robustness tests for the given model and dataset.

    This will run the model n times with different random seeds to get a distribution of the results.

    :param n_trials: number of trials to run
    :param model: model to evaluate
    :param hpam_set: hyperparameters to use
    :param path_data: path to the data directory
    :param train_dataset: training dataset
    :param test_dataset: test dataset
    :param early_stopping_dataset: early stopping dataset
    :param path_out: path to the output directory
    :param split_index: index of the split
    :param response_transformation: sklearn.preprocessing scaler like StandardScaler or MinMaxScaler to use to scale
        the target
    :param model_checkpoint_dir: directory to save model checkpoints, if "TEMPORARY": temporary directory is used
    """
    robustness_test_path = os.path.join(path_out, "robustness")
    os.makedirs(robustness_test_path, exist_ok=True)
    for trial in range(n_trials):
        print(f"Running robustness test trial {trial + 1}/{n_trials}")
        trial_file = os.path.join(
            robustness_test_path,
            f"robustness_{trial + 1}_split_{split_index}.csv",
        )
        if not os.path.isfile(trial_file):
            robustness_train_predict(
                trial=trial,
                trial_file=trial_file,
                train_dataset=train_dataset,
                test_dataset=test_dataset,
                early_stopping_dataset=early_stopping_dataset,
                model=model,
                hpam_set=hpam_set,
                path_data=path_data,
                response_transformation=response_transformation,
                model_checkpoint_dir=model_checkpoint_dir,
            )


@pipeline_function
def robustness_train_predict(
    trial: int,
    trial_file: str,
    train_dataset: DrugResponseDataset,
    test_dataset: DrugResponseDataset,
    early_stopping_dataset: DrugResponseDataset | None,
    model: DRPModel,
    hpam_set: dict,
    path_data: str,
    response_transformation: TransformerMixin | None = None,
    model_checkpoint_dir: str = "TEMPORARY",
) -> None:
    """
    Train and predict for the robustness test.

    :param trial: trial number
    :param trial_file: file to save the results to
    :param train_dataset: training dataset
    :param test_dataset: test dataset
    :param early_stopping_dataset: early stopping dataset
    :param model: model to evaluate
    :param hpam_set: hyperparameters to use
    :param path_data: path to the data directory, e.g., data/
    :param response_transformation: sklearn.preprocessing scaler like StandardScaler or MinMaxScaler to use to scale
    :param model_checkpoint_dir: directory to save model checkpoints. If "TEMPORARY", a temporary directory is created.
    """
    train_dataset.shuffle(random_state=trial)
    test_dataset.shuffle(random_state=trial)
    if early_stopping_dataset is not None:
        early_stopping_dataset.shuffle(random_state=trial)
    test_dataset = train_and_predict(
        model=model,
        hpams=hpam_set,
        path_data=path_data,
        train_dataset=train_dataset,
        prediction_dataset=test_dataset,
        early_stopping_dataset=early_stopping_dataset,
        response_transformation=response_transformation,
        model_checkpoint_dir=model_checkpoint_dir,
    )
    test_dataset.to_csv(trial_file)


def randomization_test(
    randomization_test_views: dict[str, list[str]],
    model: DRPModel,
    hpam_set: dict,
    path_data: str,
    train_dataset: DrugResponseDataset,
    test_dataset: DrugResponseDataset,
    early_stopping_dataset: DrugResponseDataset | None,
    path_out: str,
    split_index: int,
    randomization_type: str = "permutation",
    response_transformation=TransformerMixin | None,
    model_checkpoint_dir: str = "TEMPORARY",
) -> None:
    """
    Run randomization tests for the given model and dataset.

    :param randomization_test_views: views to use for the randomization tests.
        Key is the name of the randomization test and the value is a list of views to randomize
        e.g. {"randomize_genomics": ["copy_number_var", "mutation"],
        "methylation_only": ["gene_expression", "copy_number_var", "mutation"]}"
    :param model: model to evaluate
    :param hpam_set: hyperparameters to use
    :param path_data: path to the data directory
    :param train_dataset: training dataset
    :param test_dataset: test dataset
    :param early_stopping_dataset: early stopping dataset
    :param path_out: path to the output directory
    :param split_index: index of the split
    :param randomization_type: type of randomization to use. Choose from "permutation", "invariant".
        Default is "permutation" which permutes the features over the instances, keeping the
        distribution of the features the same but dissolving the relationship to the target.
        invariant randomization is done in a way that a key characteristic of the feature is preserved.
        In case of matrices, this is the mean and standard deviation of the feature view for this
        instance, for networks it is the degree distribution.
    :param response_transformation: sklearn.preprocessing scaler like StandardScaler or MinMaxScaler
        to use to scale the target
    :param model_checkpoint_dir: directory to save model checkpoints
    """
    for test_name, views in randomization_test_views.items():
        randomization_test_path = os.path.join(path_out, "randomization")
        os.makedirs(randomization_test_path, exist_ok=True)

        randomization_test_file = os.path.join(
            randomization_test_path,
            f"randomization_{test_name}_split_{split_index}.csv",
        )
        if not os.path.isfile(randomization_test_file):  # if this splits test has not been run yet
            for view in views:
                print(f"Randomizing view {view} for randomization test {test_name} ...")
                randomize_train_predict(
                    view=view,
                    test_name=test_name,
                    randomization_type=randomization_type,
                    randomization_test_file=randomization_test_file,
                    model=model,
                    hpam_set=hpam_set,
                    path_data=path_data,
                    train_dataset=train_dataset,
                    test_dataset=test_dataset,
                    early_stopping_dataset=early_stopping_dataset,
                    response_transformation=response_transformation,
                    model_checkpoint_dir=model_checkpoint_dir,
                )
        else:
            print(f"Randomization test {test_name} already exists. Skipping.")


@pipeline_function
def randomize_train_predict(
    view: str,
    test_name: str,
    randomization_type: str,
    randomization_test_file: str,
    model: DRPModel,
    hpam_set: dict,
    path_data: str,
    train_dataset: DrugResponseDataset,
    test_dataset: DrugResponseDataset,
    early_stopping_dataset: DrugResponseDataset | None,
    model_checkpoint_dir: str = "TEMPORARY",
    response_transformation: TransformerMixin | None = None,
) -> None:
    """
    Randomize the features for a given view and run the model.

    :param view: view to randomize, e.g., gene_expression
    :param test_name: name of the randomization test, e.g., SVRC_gene_expression
    :param randomization_type: type of randomization to use, e.g., permutation
    :param randomization_test_file: file to save the results to
    :param model: model to evaluate
    :param hpam_set: hyperparameters to use
    :param path_data: path to the data directory
    :param train_dataset: training dataset
    :param test_dataset: test dataset
    :param early_stopping_dataset: early stopping dataset
    :param model_checkpoint_dir: directory to save model checkpoints
    :param response_transformation: sklearn.preprocessing scaler like StandardScaler or MinMaxScaler to use to scale
    """
    cl_features, drug_features = load_features(model, path_data, train_dataset)

    # Handle case where both features are None early on
    if cl_features is None and drug_features is None:
        warnings.warn(
            "Both cl_features and drug_features are None. Skipping randomization test.",
            stacklevel=2,
        )
        return

    # Check if view is in either feature set, if not, warn and skip
    if (cl_features is not None and view not in cl_features.view_names) and (
        drug_features is not None and view not in drug_features.view_names
    ):
        warnings.warn(
            f"View {view} not found in features. Skipping randomization test {test_name} which includes this view.",
            stacklevel=2,
        )
        return

    cl_features_rand: FeatureDataset | None = None
    if cl_features is not None:
        cl_features_rand = cl_features.copy()
        cl_features_rand.randomize_features(view, randomization_type=randomization_type)  # type: ignore[union-attr]

    drug_features_rand: FeatureDataset | None = None
    if drug_features is not None:
        drug_features_rand = drug_features.copy()
        drug_features_rand.randomize_features(view, randomization_type=randomization_type)  # type: ignore[union-attr]

    test_dataset_rand = train_and_predict(
        model=model,
        hpams=hpam_set,
        path_data=path_data,
        train_dataset=train_dataset,
        prediction_dataset=test_dataset,
        early_stopping_dataset=early_stopping_dataset,
        response_transformation=response_transformation,
        cl_features=cl_features_rand,
        drug_features=drug_features_rand,
        model_checkpoint_dir=model_checkpoint_dir,
    )
    test_dataset_rand.to_csv(randomization_test_file)


def split_early_stopping(
    validation_dataset: DrugResponseDataset, test_mode: str
) -> tuple[DrugResponseDataset, DrugResponseDataset]:
    """
    Split the validation dataset into a validation and early stopping dataset.

    :param validation_dataset: validation dataset
    :param test_mode: test mode one of "LPO", "LCO", "LDO" (leave-pair-out, leave-cell-line-out, leave-drug-out)
    :returns: tuple of validation and early stopping datasets
    """
    validation_dataset.shuffle(random_state=42)
    cv_v = validation_dataset.split_dataset(
        n_cv_splits=4,
        mode=test_mode,
        split_validation=False,
        random_state=42,
    )
    # take the first fold of a 4 cv as the split ie. 3/4 for validation and 1/4 for early stopping
    validation_dataset = cv_v[0]["train"]
    early_stopping_dataset = cv_v[0]["test"]
    return validation_dataset, early_stopping_dataset


@pipeline_function
def train_and_predict(
    model: DRPModel,
    hpams: dict,
    path_data: str,
    train_dataset: DrugResponseDataset,
    prediction_dataset: DrugResponseDataset,
    early_stopping_dataset: DrugResponseDataset | None = None,
    response_transformation: TransformerMixin | None = None,
    cl_features: FeatureDataset | None = None,
    drug_features: FeatureDataset | None = None,
    model_checkpoint_dir: str = "TEMPORARY",
) -> DrugResponseDataset:
    """
    Train the model and predict the response for the prediction dataset.

    :param model: model to use, e.g., SimpleNeuralNetwork
    :param hpams: hyperparameters to use
    :param path_data: path to the data directory, e.g., data/
    :param train_dataset: training dataset
    :param prediction_dataset: prediction dataset
    :param early_stopping_dataset: early stopping dataset, optional
    :param response_transformation: normalizer to use for the response data, e.g., StandardScaler
    :param cl_features: cell line features
    :param drug_features: drug features
    :param model_checkpoint_dir: directory for model checkpoints, if "TEMPORARY", checkpoints are not saved.
            Default is "TEMPORARY"
    :returns: prediction dataset with predictions
    :raises ValueError: if train_dataset does not have a dataset_name
    """
    model.build_model(hyperparameters=hpams)
    if train_dataset.dataset_name is None:
        raise ValueError("train_dataset must have a dataset_name")
    if cl_features is None:
        print("Loading cell line features ...")
        cl_features = model.load_cell_line_features(data_path=path_data, dataset_name=train_dataset.dataset_name)
    if drug_features is None:
        print("Loading drug features ...")
        drug_features = model.load_drug_features(data_path=path_data, dataset_name=train_dataset.dataset_name)

    cell_lines_to_keep = cl_features.identifiers if cl_features is not None else None
    drugs_to_keep = drug_features.identifiers if drug_features is not None else None

    # making sure there are no missing features:
    len_train_before = len(train_dataset)
    len_pred_before = len(prediction_dataset)
    print(f"Number of cell lines in features: {len(cell_lines_to_keep)}")
    if drugs_to_keep is not None:
        print(f"Number of drugs in features: {len(drugs_to_keep)}")
    print(f"Number of cell lines in train dataset: {len(np.unique(train_dataset.cell_line_ids))}")
    print(f"Number of drugs in train dataset: {len(np.unique(train_dataset.drug_ids))}")

    train_dataset.reduce_to(cell_line_ids=cell_lines_to_keep, drug_ids=drugs_to_keep)
    prediction_dataset.reduce_to(cell_line_ids=cell_lines_to_keep, drug_ids=drugs_to_keep)
    print(f"Reduced training dataset from {len_train_before} to {len(train_dataset)}, because of missing features")
    print(
        f"Reduced prediction dataset from {len_pred_before} to {len(prediction_dataset)}, because of missing features"
    )

    if early_stopping_dataset is not None:
        len_es_before = len(early_stopping_dataset)
        early_stopping_dataset.reduce_to(cell_line_ids=cell_lines_to_keep, drug_ids=drugs_to_keep)
        print(f"Reduced early stopping dataset from {len_es_before} to {len(early_stopping_dataset)}")

    if response_transformation:
        train_dataset.fit_transform(response_transformation)
        if early_stopping_dataset is not None:
            early_stopping_dataset.transform(response_transformation)
        prediction_dataset.transform(response_transformation)

    print("Training model ...")
    if model_checkpoint_dir == "TEMPORARY":
        with tempfile.TemporaryDirectory() as temp_dir:
            print(f"Using temporary directory: {temp_dir} for model checkpoints")
            model.train(
                output=train_dataset,
                output_earlystopping=early_stopping_dataset,
                cell_line_input=cl_features,
                drug_input=drug_features,
                model_checkpoint_dir=temp_dir,
            )
    else:
        if not os.path.exists(model_checkpoint_dir):
            os.makedirs(model_checkpoint_dir, exist_ok=True)
        print(f"Using directory: {model_checkpoint_dir} for model checkpoints")
        model.train(
            output=train_dataset,
            output_earlystopping=early_stopping_dataset,
            cell_line_input=cl_features,
            drug_input=drug_features,
            model_checkpoint_dir=model_checkpoint_dir,
        )

    if len(prediction_dataset) > 0:
        prediction_dataset._predictions = model.predict(
            cell_line_ids=prediction_dataset.cell_line_ids,
            drug_ids=prediction_dataset.drug_ids,
            cell_line_input=cl_features,
            drug_input=drug_features,
        )

        if response_transformation:
            prediction_dataset.inverse_transform(response_transformation)
    else:
        prediction_dataset._predictions = np.array([])

    return prediction_dataset


def train_and_evaluate(
    model: DRPModel,
    hpams: dict[str, list],
    path_data: str,
    train_dataset: DrugResponseDataset,
    validation_dataset: DrugResponseDataset,
    early_stopping_dataset: DrugResponseDataset | None = None,
    response_transformation: TransformerMixin | None = None,
    metric: str = "rmse",
    model_checkpoint_dir: str = "TEMPORARY",
) -> dict[str, float]:
    """
    Train and evaluate the model, i.e., call train_and_predict() and then evaluate().

    :param model: model to use
    :param hpams: hyperparameters to use
    :param path_data: path to the data directory
    :param train_dataset: training dataset
    :param validation_dataset: validation dataset
    :param early_stopping_dataset: early stopping dataset
    :param response_transformation: normalizer to use for the response data
    :param metric: metric to evaluate the model on
    :param model_checkpoint_dir: directory to save model checkpoints
    :returns: dictionary of the evaluation results, e.g., {"RMSE": 0.1}
    """
    validation_dataset = train_and_predict(
        model=model,
        hpams=hpams,
        path_data=path_data,
        train_dataset=train_dataset,
        prediction_dataset=validation_dataset,
        early_stopping_dataset=early_stopping_dataset,
        response_transformation=response_transformation,
        model_checkpoint_dir=model_checkpoint_dir,
    )
    return evaluate(validation_dataset, metric=[metric])


def hpam_tune(
    model: DRPModel,
    train_dataset: DrugResponseDataset,
    validation_dataset: DrugResponseDataset,
    hpam_set: list[dict],
    early_stopping_dataset: DrugResponseDataset | None = None,
    response_transformation: TransformerMixin | None = None,
    metric: str = "RMSE",
    path_data: str = "data",
    model_checkpoint_dir: str = "TEMPORARY",
) -> dict:
    """
    Tune the hyperparameters for the given model in an iterative manner.

    :param model: model to use
    :param train_dataset: training dataset
    :param validation_dataset: validation dataset
    :param hpam_set: hyperparameters to tune
    :param early_stopping_dataset: early stopping dataset
    :param response_transformation: normalizer to use for the response data
    :param metric: metric to evaluate which model is the best
    :param path_data: path to the data directory, e.g., data/
    :param model_checkpoint_dir: directory to save model checkpoints
    :returns: best hyperparameters
    :raises AssertionError: if hpam_set is empty
    """
    if len(hpam_set) == 0:
        raise AssertionError("hpam_set must contain at least one hyperparameter configuration")
    if len(hpam_set) == 1:
        return hpam_set[0]

    best_hyperparameters = None
    mode = get_mode(metric)
    best_score = float("inf") if mode == "min" else float("-inf")
    for hyperparameter in hpam_set:
        print(f"Training model with hyperparameters: {hyperparameter}")
        score = train_and_evaluate(
            model=model,
            hpams=hyperparameter,
            path_data=path_data,
            train_dataset=train_dataset,
            validation_dataset=validation_dataset,
            early_stopping_dataset=early_stopping_dataset,
            metric=metric,
            response_transformation=response_transformation,
            model_checkpoint_dir=model_checkpoint_dir,
        )[metric]

        if np.isnan(score):
            continue

        if (mode == "min" and score < best_score) or (mode == "max" and score > best_score):
            print(f"current best {metric} score: {np.round(score, 3)}")
            best_score = score
            best_hyperparameters = hyperparameter

    if best_hyperparameters is None:
        warnings.warn("all hpams lead to NaN respone. using last hpam combination.", stacklevel=2)
        best_hyperparameters = hyperparameter

    return best_hyperparameters


def hpam_tune_raytune(
    model: DRPModel,
    train_dataset: DrugResponseDataset,
    validation_dataset: DrugResponseDataset,
    early_stopping_dataset: DrugResponseDataset | None,
    hpam_set: list[dict],
    response_transformation: TransformerMixin | None = None,
    metric: str = "RMSE",
    ray_path: str = "raytune",
    path_data: str = "data",
    model_checkpoint_dir: str = "TEMPORARY",
) -> dict:
    """
    Tune the hyperparameters for the given model using raytune. This requires ray to be installed.

    :param model: model to use
    :param train_dataset: training dataset
    :param validation_dataset: validation dataset
    :param early_stopping_dataset: early stopping dataset
    :param hpam_set: hyperparameters to tune
    :param response_transformation: normalizer to use for the response data
    :param metric: metric to evaluate which model is the best
    :param ray_path: path to the raytune directory
    :param path_data: path to the data directory, e.g., data/
    :param model_checkpoint_dir: directory to save model checkpoints
    :returns: best hyperparameters
    """
    if len(hpam_set) == 1:
        return hpam_set[0]
    ray.init(_temp_dir=os.path.join(os.path.expanduser("~"), "raytmp"))
    if torch.cuda.is_available():
        resources_per_trial = {"gpu": 1}  # TODO make this user defined
    else:
        resources_per_trial = {"cpu": 1}  # TODO make this user defined
    analysis = ray.tune.run(
        lambda hpams: train_and_evaluate(
            model=model,
            hpams=hpams,
            path_data=path_data,
            train_dataset=train_dataset,
            validation_dataset=validation_dataset,
            early_stopping_dataset=early_stopping_dataset,
            metric=metric,
            response_transformation=response_transformation,
            model_checkpoint_dir=model_checkpoint_dir,
        ),
        config=ray.tune.grid_search(hpam_set),
        mode="min",
        num_samples=5,
        resources_per_trial=resources_per_trial,
        chdir_to_trial_dir=False,
        verbose=0,
        storage_path=ray_path,
    )

    mode = get_mode(metric)
    best_config = analysis.get_best_config(metric=metric, mode=mode)
    return best_config


@pipeline_function
def make_model_list(models: list[type[DRPModel]], response_data: DrugResponseDataset) -> dict[str, str]:
    """
    Make a list of models to evaluate: if it is a single drug model, add the drug id to the model name.

    :param models: list of models to evaluate
    :param response_data: response data, needed to get the unique drugs for single drug models
    :returns: dictionary of model names: model class, e.g., {"SimpleNeuralNetwork": "SimpleNeuralNetwork",
        "MOLIR.Afatinib": "MOLIR"}
    """
    model_list = {}
    unique_drugs = np.unique(response_data.drug_ids)
    for model in models:
        if model.is_single_drug_model:
            for drug in unique_drugs:
                model_list[f"{model.get_model_name()}.{drug}"] = model.get_model_name()
        else:
            model_list[model.get_model_name()] = model.get_model_name()
    return model_list


@pipeline_function
def get_model_name_and_drug_id(model_name: str) -> tuple[str, str | None]:
    """
    Get the model name and drug id from the model name.

    :param model_name: model name, e.g., SimpleNeuralNetwork or MOLIR.Afatinib
    :returns: tuple of model name and, potentially drug id if it is a single drug model
    :raises AssertionError: if the model name is not found in the model factory
    """
    if model_name in MULTI_DRUG_MODEL_FACTORY:
        return model_name, None
    else:
        name_split = model_name.split(".")
        model_name = name_split[0]
        if model_name not in SINGLE_DRUG_MODEL_FACTORY:
            raise AssertionError(
                f"Model {model_name} not found in MODEL_FACTORY or SINGLE_DRUG_MODEL_FACTORY. "
                "Please add the model to the factory."
            )
        drug_id = name_split[1]

        return model_name, drug_id


@pipeline_function
def get_datasets_from_cv_split(
    split: dict[str, DrugResponseDataset], model_class: type[DRPModel], model_name: str, drug_id: str | None = None
) -> tuple[DrugResponseDataset, DrugResponseDataset, DrugResponseDataset | None, DrugResponseDataset]:
    """
    Get train, validation, (early stopping), and test datasets from the CV split.

    :param split: dictionary of the CV split
    :param model_class: model class
    :param model_name: model name
    :param drug_id: drug id for single drug models
    :returns: tuple of train, validation, (early stopping), and test datasets
    """
    train_dataset = split["train"]
    validation_dataset = split["validation"]
    test_dataset = split["test"]

    if model_class.early_stopping:
        validation_dataset = split["validation_es"]
        early_stopping_dataset = split["early_stopping"]
    else:
        early_stopping_dataset = None

    if model_name in SINGLE_DRUG_MODEL_FACTORY.keys():
        output_mask = train_dataset.drug_ids == drug_id
        train_cp = train_dataset.copy()
        train_cp.mask(output_mask)
        validation_mask = validation_dataset.drug_ids == drug_id
        val_cp = validation_dataset.copy()
        val_cp.mask(validation_mask)
        test_mask = test_dataset.drug_ids == drug_id
        test_cp = test_dataset.copy()
        test_cp.mask(test_mask)
        if early_stopping_dataset is not None:
            es_mask = early_stopping_dataset.drug_ids == drug_id
            es_cp = early_stopping_dataset.copy()
            es_cp.mask(es_mask)
            return train_cp, val_cp, es_cp, test_cp
        return train_cp, val_cp, None, test_cp

    return (
        train_dataset,
        validation_dataset,
        early_stopping_dataset,
        test_dataset,
    )


@pipeline_function
def generate_data_saving_path(model_name, drug_id, result_path, suffix) -> str:
    """
    Generate a path to save data to.

    For single drug models, the path is result_path/model_name/drugs/drug_id/suffix.
    For all others, it is result_path/model_name/suffix.
    :param model_name: model name
    :param drug_id: drug id
    :param result_path: path to the results directory
    :param suffix: suffix to add to the path, e.g., "predictions", "best_hpams", "randomization", "robustness"
    :returns: path to save data to
    """
    is_single_drug_model = model_name in SINGLE_DRUG_MODEL_FACTORY
    if is_single_drug_model:
        model_path = os.path.join(result_path, model_name, "drugs", drug_id, suffix)
    else:
        model_path = os.path.join(result_path, model_name, suffix)
    os.makedirs(model_path, exist_ok=True)
    return model_path



"""Contains a custom MultiHeadAttentionLayer for the DIPK model."""

import torch
from torch import nn


class MultiHeadAttentionLayer(nn.Module):
    """Custom multi-head attention layer for the DIPK model."""

    def __init__(self, hid_dim: int, n_heads: int, dropout: float, device: str | torch.device | int | None):
        """
        Initialize the multi-head attention layer.

        :param hid_dim: dimension of hidden layer
        :param n_heads: number of heads
        :param dropout: dropout rate
        :param device: which device to use, e.g. "cuda" or "cpu"
        :raises ValueError: if hidden dimension is not divisible by the number of heads
        """
        super().__init__()

        # Ensure head dimension divides evenly
        if hid_dim % n_heads != 0:
            raise ValueError("Hidden dimension must be divisible by the number of heads.")

        # Define dimensions
        self.hid_dim = hid_dim
        self.n_heads = n_heads
        self.head_dim = hid_dim // n_heads

        # Define fully connected layers for Q, K, V, and output
        self.fc_q = nn.Linear(hid_dim, hid_dim)
        self.fc_k = nn.Linear(hid_dim, hid_dim)
        self.fc_v = nn.Linear(hid_dim, hid_dim)
        self.fc_o = nn.Linear(hid_dim, hid_dim)

        # Dropout and scaling factor
        self.dropout = nn.Dropout(dropout)
        self.scale = torch.sqrt(torch.tensor(self.head_dim, dtype=torch.float32, device=device))

    def forward(
        self, query: torch.Tensor, key: torch.Tensor, value: torch.Tensor, mask: torch.Tensor | None = None
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """
        Forward pass of the multi-head attention layer.

        :param query: query tensor
        :param key: key tensor
        :param value: value tensor
        :param mask: mask tensor
        :returns: output tensor and attention tensor
        """
        batch_size = query.size(0)

        # Transform inputs
        transformed_query = self.fc_q(query)
        transformed_key = self.fc_k(key)
        transformed_value = self.fc_v(value)

        # Split into heads and transpose for multi-head processing
        transformed_query = transformed_query.view(batch_size, -1, self.n_heads, self.head_dim).permute(0, 2, 1, 3)
        transformed_key = transformed_key.view(batch_size, -1, self.n_heads, self.head_dim).permute(0, 2, 1, 3)
        transformed_value = transformed_value.view(batch_size, -1, self.n_heads, self.head_dim).permute(0, 2, 1, 3)

        # Scaled dot-product attention
        energy = torch.matmul(transformed_query, transformed_key.transpose(-2, -1)) / self.scale
        if mask is not None:
            energy = energy.masked_fill(mask == 0, float("-inf"))
        attention = torch.softmax(energy, dim=-1)

        # Apply attention weights
        x = torch.matmul(self.dropout(attention), transformed_value)

        # Concatenate heads and pass through output layer
        x = x.permute(0, 2, 1, 3).contiguous().view(batch_size, -1, self.hid_dim)
        x = self.fc_o(x)

        return x, attention



"""
Contains the SRMF (Similarity Regularization Matrix Factorization) model.

Original publication: Wang, L., Li, X., Zhang, L. et al. Improved anticancer drug response prediction in cell lines
using matrix factorization with similarity regularization. BMC Cancer 17, 513 (2017).
https://doi.org/10.1186/s12885-017-3500-5.
Matlab code adapted from https://github.com/linwang1982/SRMF.
"""

import numpy as np
import pandas as pd
from scipy.spatial.distance import jaccard

from drevalpy.datasets.dataset import DrugResponseDataset, FeatureDataset
from drevalpy.models.drp_model import DRPModel
from drevalpy.models.utils import load_and_reduce_gene_features, load_drug_fingerprint_features


class SRMF(DRPModel):
    """
    SRMF model: Similarity Regularization Matrix Factorization.

    The primary idea is to map m drugs and n cell lines into a shared latent space, with a low dimensionality K,
    where :math:`K << min (m, n)`. The properties of a drug :math:`d_i` and a cell line :math:`c_j` are described by
    two latent coordinates :math:`u_i` and :math:`v_j` (K dimensional row vectors), respectively. The drug response
    matrix Y is approximated by: :math:`min_{U,V} || W * (Y - U * V^T) ||^2_F + lambda_l * (||U||^2_F + ||V||^2_F) +
    lambda_d * ||S_d - U * U^T||^2_F + lambda_c * ||S_c - V * V^T||^2_F`
    where W is a weight matrix (:math:`W_{ij} = 1 if Y_{ij}` is a known response value, else 0). U, V contain
    :math:`u_i`, :math:`v_j` as row vectors, respectively, :math:`||.||_F` is the Frobenius norm. To avoid overfitting,
    L2 regularization is used. :math:`S_d, S_c` are drug/cell line similarity matrices. Differences between two
    drugs/cell lines are minimized in latent space.
    """

    cell_line_views = ["gene_expression"]
    drug_views = ["fingerprints"]

    def __init__(self) -> None:
        """Initalization method for SRMF Model."""
        super().__init__()
        self.best_u: pd.DataFrame = pd.DataFrame()
        self.best_v: pd.DataFrame = pd.DataFrame()
        self.w: pd.DataFrame = pd.DataFrame()
        self.k: int = 45
        self.lambda_l: float = 0.01
        self.lambda_d: float = 0.0
        self.lambda_c: float = 0.01
        self.max_iter: int = 50
        self.seed: int = 1

    @classmethod
    def get_model_name(cls) -> str:
        """
        Returns the model name.

        :returns: SRMF
        """
        return "SRMF"

    def build_model(self, hyperparameters: dict) -> None:
        """
        Initializes hyperparameters for SRMF model.

        K is the latent dimensionality, lambda_l, lambda_d, lambda_c are regularization parameters, max_iter is the
        number of iterations, seed is the random seed.

        :param hyperparameters: dictionary containing the hyperparameters
        """
        self.k = hyperparameters.get("K", 45)
        self.lambda_l = hyperparameters.get("lambda_l", 0.01)
        self.lambda_d = hyperparameters.get("lambda_d", 0)
        self.lambda_c = hyperparameters.get("lambda_c", 0.01)
        self.max_iter = hyperparameters.get("max_iter", 50)
        self.seed = hyperparameters.get("seed", 1)

    def train(
        self,
        output: DrugResponseDataset,
        cell_line_input: FeatureDataset,
        drug_input: FeatureDataset | None = None,
        output_earlystopping: DrugResponseDataset | None = None,
        model_checkpoint_dir: str = "checkpoints",
    ) -> None:
        """
        Prepares data and trains the SRMF model.

        :param output: response data
        :param cell_line_input: feature data for cell lines
        :param drug_input: feature data for drugs
        :param output_earlystopping: optional early stopping dataset, not used in SRMF
        :param model_checkpoint_dir: directory to save the model checkpoints, not used in SRMF
        :raises ValueError: if drug_input is None
        """
        if drug_input is None:
            raise ValueError("SRMF requires drug features.")

        drugs = np.unique(drug_input.identifiers)  # transductive approach - all drug features are used
        cell_lines = np.unique(cell_line_input.identifiers)  # transductive approach - all cell line features are used

        drug_features = drug_input.features

        drug_similarity = pd.DataFrame(index=drugs, columns=drugs, dtype=float)  # jaccard similarity

        for drug_from in drugs:
            for drug_to in drugs:
                # skip if already computed
                if not np.isnan(drug_similarity.loc[drug_from, drug_to]):
                    continue
                drug_similarity.loc[drug_from, drug_to] = 1 - jaccard(
                    drug_features[drug_from]["fingerprints"],
                    drug_features[drug_to]["fingerprints"],
                )
                drug_similarity.loc[drug_to, drug_from] = drug_similarity.loc[drug_from, drug_to]

        cell_line_features = cell_line_input.get_feature_matrix(view="gene_expression", identifiers=cell_lines)
        # pearson correlation as similarity
        cell_line_similarity = np.corrcoef(cell_line_features, rowvar=True)

        # Prepare response and weight matrices
        drug_response_matrix = output.to_dataframe()
        drug_response_matrix = drug_response_matrix.groupby(["cell_line_id", "drug_id"]).mean().reset_index()
        drug_response_matrix = drug_response_matrix.pivot(index="cell_line_id", columns="drug_id", values="response")

        drug_response_matrix = drug_response_matrix.reindex(
            index=cell_lines, columns=drugs
        )  # missing rows and columns are filled with NaN

        self.w = ~np.isnan(drug_response_matrix)
        drug_response_matrix = drug_response_matrix.copy()
        drug_response_matrix[np.isnan(drug_response_matrix)] = 0

        # Train the model
        best_u, best_v = self._cmf(
            w=self.w.T.values,
            int_mat=drug_response_matrix.values.T,
            drug_mat=drug_similarity.values,
            cell_mat=cell_line_similarity,
        )
        self.best_u = pd.DataFrame(best_u, index=drugs)
        self.best_v = pd.DataFrame(best_v, index=cell_lines)

    def predict(
        self,
        cell_line_ids: np.ndarray,
        drug_ids: np.ndarray,
        cell_line_input: FeatureDataset,
        drug_input: FeatureDataset | None = None,
    ) -> np.ndarray:
        """
        Predicts the drug response based on the trained latent factors.

        :param drug_ids: drug identifiers
        :param cell_line_ids: cell line identifiers
        :param cell_line_input: not needed for prediction in SRMF
        :param drug_input: not needed for prediction in SRMF
        :returns: predicted response matrix
        """
        best_u = self.best_u.loc[drug_ids].values
        best_v = self.best_v.loc[cell_line_ids].values
        # calculate the diagonal of the matrix product which is the prediction,
        # faster than np.dot(best_u, best_v.T).diagonal()
        diagonal_predictions = np.einsum("ij,ji->i", best_u, best_v.T)

        return diagonal_predictions

    def _cmf(self, w, int_mat, drug_mat, cell_mat) -> tuple[np.ndarray, np.ndarray]:
        """
        Implements the SRMF model with specific update rules and regularization.

        :param w: weight matrix
        :param int_mat: interaction matrix
        :param drug_mat: drug similarity matrix
        :param cell_mat: cell line similarity matrix
        :returns: best drug and cell line latent factors
        """
        np.random.seed(self.seed)
        m, n = w.shape
        u0 = np.sqrt(1 / self.k) * np.random.randn(m, self.k)
        v0 = np.sqrt(1 / self.k) * np.random.randn(n, self.k)

        best_u, best_v = u0, v0

        last_loss = self._compute_loss(u0, v0, w, int_mat, drug_mat, cell_mat)
        best_loss = last_loss
        wr = w * int_mat

        for _ in range(self.max_iter):
            u = self._alg_update(u0, v0, w, wr, drug_mat, self.lambda_l, self.lambda_d)
            v = self._alg_update(v0, u, w.T, wr.T, cell_mat, self.lambda_l, self.lambda_c)
            curr_loss = self._compute_loss(u, v, w, int_mat, drug_mat, cell_mat)

            if curr_loss < best_loss:
                best_u, best_v = u, v
                best_loss = curr_loss

            delta_loss = (curr_loss - last_loss) / last_loss
            if abs(delta_loss) < 1e-6:
                break

            last_loss = curr_loss
            u0, v0 = u, v

        return best_u, best_v

    def _compute_loss(self, u, v, w, int_mat, drug_mat, cell_mat) -> np.float64:
        """
        Computes the loss for SRMF, including similarity regularization.

        :param u: drug latent factors
        :param v: cell line latent factors
        :param w: weight matrix
        :param int_mat: interaction matrix
        :param drug_mat: drug similarity matrix
        :param cell_mat: cell line similarity matrix
        :returns: loss value
        """
        loss = np.sum((w * (int_mat - np.dot(u, v.T))) ** 2)
        loss += self.lambda_l * (np.sum(u**2) + np.sum(v**2))
        loss += self.lambda_d * np.sum((drug_mat - np.dot(u, u.T)) ** 2)
        loss += self.lambda_c * np.sum((cell_mat - np.dot(v, v.T)) ** 2)
        return loss

    def _alg_update(self, u, v, w, r, s, lambda_l, lambda_d) -> np.ndarray:
        """
        Algorithm update rule for u or v in the SRMF model.

        :param u: drug latent factors
        :param v: cell line latent factors
        :param w: weight matrix
        :param r: weight * interaction matrix
        :param s: drug/cell line similarity matrix
        :param lambda_l: regularization parameter
        :param lambda_d: drug/cell line similarity regularization parameter
        :returns: updated u or v
        """
        x = np.dot(r, v) + 2 * lambda_d * np.dot(s, u)
        y = 2 * lambda_d * np.dot(u.T, u)
        u0 = np.zeros_like(u)
        d = np.dot(v.T, v)
        m, _ = w.shape

        for i in range(m):
            ii = np.where(w[i, :] > 0)[0]
            if ii.size == 0:
                b = y + lambda_l * np.eye(u.shape[1])
            elif ii.size == w.shape[1]:
                b = d + y + lambda_l * np.eye(u.shape[1])
            else:
                a = np.dot(v[ii, :].T, v[ii, :])
                b = a + y + lambda_l * np.eye(u.shape[1])

            u0[i, :] = np.linalg.solve(b, x[i, :])

        return u0

    def load_cell_line_features(self, data_path: str, dataset_name: str) -> FeatureDataset:
        """
        Loads the cell line features, in this case the gene expression features.

        :param data_path: Path to the gene expression and landmark genes, e.g., data/
        :param dataset_name: Name of the dataset, e.g., GDSC2
        :returns: FeatureDataset containing the cell line gene expression features, filtered
            through the landmark genes
        """
        return load_and_reduce_gene_features(
            feature_type="gene_expression",
            gene_list=None,
            data_path=data_path,
            dataset_name=dataset_name,
        )

    def load_drug_features(self, data_path: str, dataset_name: str) -> FeatureDataset:
        """
        Loads the drug features, in this case the drug fingerprints.

        :param data_path: Path to the drug features, in this case the drug fingerprints, e.g., data/
        :param dataset_name: Name of the dataset, e.g., GDSC2
        :returns: FeatureDataset containing the drug fingerprint features
        """
        return load_drug_fingerprint_features(data_path, dataset_name)



"""
Contains the MOLIR model, a regression adaptation of the MOLI model.

Original authors: Sharifi-Noghabi et al. (2019, 10.1093/bioinformatics/btz318)
Code adapted from their Github: https://github.com/hosseinshn/MOLI
and Hauptmann et al. (2023, 10.1186/s12859-023-05166-7) https://github.com/kramerlab/Multi-Omics_analysis
"""

from typing import Any

import numpy as np
from sklearn.feature_selection import VarianceThreshold
from sklearn.preprocessing import StandardScaler

from ...datasets.dataset import DrugResponseDataset, FeatureDataset
from ..drp_model import DRPModel
from ..utils import get_multiomics_feature_dataset
from .utils import MOLIModel, get_dimensions_of_omics_data


class MOLIR(DRPModel):
    """
    Regression extension of MOLI: multi-omics late integration deep neural network.

    Takes somatic mutation, copy number variation and gene expression data as input. MOLI uses type-specific encoding
    subnetworks to learn features for each omics type, concatenates them into one representation and optimizes this
    representation via a combined cost function consisting of a triplet loss and a binary cross-entropy loss.
    We use a regression adaption with MSE loss and a mechanism to find positive and negative samples.
    """

    is_single_drug_model = True
    cell_line_views = ["gene_expression", "mutations", "copy_number_variation_gistic"]
    drug_views = []
    early_stopping = True

    def __init__(self) -> None:
        """
        Initializes the MOLIR model.

        The hyperparameters are set in build_model, the model is set in train when we know the dimensionality of the
        gene expression, mutation and copy number variation data.
        """
        super().__init__()
        self.model: MOLIModel | None = None
        self.hyperparameters: dict[str, Any] = dict()

    @classmethod
    def get_model_name(cls) -> str:
        """
        Returns the model name.

        :returns: MOLIR
        """
        return "MOLIR"

    def build_model(self, hyperparameters: dict[str, Any]) -> None:
        """
        Builds the model from hyperparameters.

        :param hyperparameters: Custom hyperparameters for the model, includes mini_batch, layer dimensions (h_dim1,
            h_dim2, h_dim3), learning_rate, dropout_rate, weight_decay, gamma, epochs, and margin.
        """
        self.hyperparameters = hyperparameters

    def train(
        self,
        output: DrugResponseDataset,
        cell_line_input: FeatureDataset,
        drug_input: FeatureDataset | None = None,
        output_earlystopping: DrugResponseDataset | None = None,
        model_checkpoint_dir: str = "checkpoints",
    ) -> None:
        """
        Initializes and trains the model.

        First, the gene expression data is reduced using a variance threshold (0.05) and standardized. Then,
        the model is initialized with the hyperparameters and the dimensions of the gene expression, mutation and
        copy number variation data. If there is no training data, the model is set to None (and predictions will be
        skipped as well). If there is not enough training data, the predictions will be made on the randomly
        initialized model.

        :param output: drug response data
        :param cell_line_input: cell line omics features, i.e., gene expression, mutations and copy number variation
        :param drug_input: drug features, not needed
        :param output_earlystopping: early stopping data, not used when there is not enough data
        :param model_checkpoint_dir: directory to save the model checkpoints
        """
        if len(output) > 0:
            selector_gex = VarianceThreshold(0.05)
            cell_line_input.fit_transform_features(
                train_ids=np.unique(output.cell_line_ids),
                transformer=selector_gex,
                view="gene_expression",
            )
            scaler_gex = StandardScaler()
            cell_line_input.fit_transform_features(
                train_ids=np.unique(output.cell_line_ids),
                transformer=scaler_gex,
                view="gene_expression",
            )
            if output_earlystopping is not None and self.early_stopping and len(output_earlystopping) < 2:
                output_earlystopping = None
            dim_gex, dim_mut, dim_cnv = get_dimensions_of_omics_data(cell_line_input)
            self.model = MOLIModel(
                hpams=self.hyperparameters,
                input_dim_expr=dim_gex,
                input_dim_mut=dim_mut,
                input_dim_cnv=dim_cnv,
            )
            if len(output) >= self.hyperparameters["mini_batch"]:
                self.model.fit(
                    output_train=output,
                    cell_line_input=cell_line_input,
                    output_earlystopping=output_earlystopping,
                    model_checkpoint_dir=model_checkpoint_dir,
                )
            else:
                print(f"Not enough training data provided ({len(output)}), will predict on randomly initialized model.")
        else:
            print("No training data provided, skipping model")
            self.model = None

    def predict(
        self,
        cell_line_ids: np.ndarray,
        drug_ids: np.ndarray,
        cell_line_input: FeatureDataset,
        drug_input: FeatureDataset | None = None,
    ) -> np.ndarray:
        """
        Predicts the drug response.

        If there was no training data, only nans will be returned.

        :param cell_line_ids: Cell lines to predict
        :param drug_ids: Drugs to predict
        :param cell_line_input: cell line omics features
        :param drug_input: drug features, not needed
        :returns: Predicted drug response
        """
        input_data = self.get_feature_matrices(
            cell_line_ids=cell_line_ids,
            drug_ids=drug_ids,
            cell_line_input=cell_line_input,
            drug_input=drug_input,
        )
        (gene_expression, mutations, cnvs) = (
            input_data["gene_expression"],
            input_data["mutations"],
            input_data["copy_number_variation_gistic"],
        )
        if self.model is None:
            print("No model trained, will predict NA.")
            return np.array([np.nan] * len(cell_line_ids))
        return self.model.predict(gene_expression, mutations, cnvs)

    def load_cell_line_features(self, data_path: str, dataset_name: str) -> FeatureDataset:
        """
        Loads the cell line features: gene expression, mutations and copy number variation.

        :param data_path: path to the data
        :param dataset_name: name of the dataset
        :returns: FeatureDataset with gene expression, mutations and copy number variation
        """
        feature_dataset = get_multiomics_feature_dataset(
            data_path=data_path,
            dataset_name=dataset_name,
            gene_list=None,
            omics=self.cell_line_views,
        )
        # log transformation
        feature_dataset.apply(function=np.log, view="gene_expression")
        return feature_dataset

    def load_drug_features(self, data_path: str, dataset_name: str) -> FeatureDataset | None:
        """
        Returns None, as drug features are not needed for MOLIR.

        :param data_path: path to the data
        :param dataset_name: name of the dataset
        :returns: None
        """
        return None



"""Utility functions for the visualization part of the package."""

import os
import pathlib
import re
import shutil
from typing import TextIO

import importlib_resources
import pandas as pd

from ..datasets.dataset import DrugResponseDataset
from ..evaluation import AVAILABLE_METRICS, evaluate
from ..pipeline_function import pipeline_function
from .corr_comp_scatter import CorrelationComparisonScatter
from .critical_difference_plot import CriticalDifferencePlot
from .html_tables import HTMLTable
from .regression_slider_plot import RegressionSliderPlot
from .vioheat import VioHeat


def _parse_layout(f: TextIO, path_to_layout: str) -> None:
    """
    Parse the layout file and write it to the output file.

    :param f: file to write to
    :param path_to_layout: path to the layout file
    """
    with open(path_to_layout, encoding="utf-8") as layout_f:
        layout = layout_f.readlines()
    if path_to_layout.endswith("index_layout.html"):
        # remove the last 2 lines (</body>, </html>)
        layout = layout[:-2]
    else:
        # remove the last 3 lines (</div>, </body>, </html>)
        layout = layout[:-3]
    f.write("".join(layout))


def parse_results(path_to_results: str) -> tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """
    Parse the results from the given directory.

    :param path_to_results: path to the results directory
    :returns: evaluation results, evaluation results per drug, evaluation results per cell line, and true vs. predicted
        values
    """
    print("Generating result tables ...")
    # generate list of all result files
    result_dir = pathlib.Path(path_to_results)
    result_files = list(result_dir.rglob("*.csv"))
    # filter for all files that follow this pattern:
    # result_dir/*/{predictions|cross_study|randomization|robustness}/*.csv
    # Convert the path to a forward-slash version for the regex (for Windows)
    result_dir_str = str(result_dir).replace("\\", "/")
    pattern = re.compile(
        rf"{result_dir_str}/(LPO|LCO|LDO)/[^/]+/(predictions|cross_study|randomization|robustness)/.*\.csv$"
    )
    result_files = [file for file in result_files if pattern.match(str(file).replace("\\", "/"))]

    # inititalize dictionaries to store the evaluation results
    evaluation_results = None
    evaluation_results_per_drug = None
    evaluation_results_per_cell_line = None
    true_vs_pred = None

    # read every result file and compute the evaluation metrics
    for file in result_files:
        rel_file = str(os.path.normpath(file.relative_to(result_dir))).replace("\\", "/")
        print(f'Evaluating file: "{rel_file}" ...')
        file_parts = rel_file.split("/")
        lpo_lco_ldo = file_parts[0]
        algorithm = file_parts[1]
        (
            overall_eval,
            eval_results_per_drug,
            eval_results_per_cl,
            t_vs_p,
            model_name,
        ) = evaluate_file(pred_file=file, test_mode=lpo_lco_ldo, model_name=algorithm)

        evaluation_results = (
            overall_eval if evaluation_results is None else pd.concat([evaluation_results, overall_eval])
        )
        true_vs_pred = t_vs_p if true_vs_pred is None else pd.concat([true_vs_pred, t_vs_p])

        if eval_results_per_drug is not None:
            evaluation_results_per_drug = (
                eval_results_per_drug
                if evaluation_results_per_drug is None
                else pd.concat([evaluation_results_per_drug, eval_results_per_drug])
            )

        if eval_results_per_cl is not None:
            evaluation_results_per_cell_line = (
                eval_results_per_cl
                if evaluation_results_per_cell_line is None
                else pd.concat([evaluation_results_per_cell_line, eval_results_per_cl])
            )

    return (
        evaluation_results,
        evaluation_results_per_drug,
        evaluation_results_per_cell_line,
        true_vs_pred,
    )


@pipeline_function
def evaluate_file(
    pred_file: pathlib.Path, test_mode: str, model_name: str
) -> tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame, str]:
    """
    Evaluate the predictions from the final models.

    :param pred_file: path to the prediction file
    :param test_mode: test mode, e.g., LPO
    :param model_name: model name, e.g., SimpleNeuralNetwork
    :return: evaluation results, evaluation results per drug, evaluation results per cell line, true vs. predicted
        values, and model name
    """
    print("Parsing file:", os.path.normpath(pred_file))
    dataset = DrugResponseDataset.from_csv(pred_file)

    model = _generate_model_names(test_mode=test_mode, model_name=model_name, pred_file=pred_file)

    # overall evaluation
    overall_eval = {model: evaluate(dataset, list(AVAILABLE_METRICS.keys()))}

    true_vs_pred = pd.DataFrame(
        {
            "model": [model for _ in range(len(dataset.response))],
            "drug": dataset.drug_ids,
            "cell_line": dataset.cell_line_ids,
            "y_true": dataset.response,
            "y_pred": dataset.predictions,
        }
    )

    evaluation_results_per_drug = None
    evaluation_results_per_cl = None
    norm_drug_eval_results: dict[str, dict[str, float]] = {}
    norm_cl_eval_results: dict[str, dict[str, float]] = {}

    if "LPO" in model or "LCO" in model:
        norm_drug_eval_results, evaluation_results_per_drug = _evaluate_per_group(
            df=true_vs_pred,
            group_by="drug",
            norm_group_eval_results=norm_drug_eval_results,
            eval_results_per_group=evaluation_results_per_drug,
            model=model,
        )
    if "LPO" in model or "LDO" in model:
        norm_cl_eval_results, evaluation_results_per_cl = _evaluate_per_group(
            df=true_vs_pred,
            group_by="cell_line",
            norm_group_eval_results=norm_cl_eval_results,
            eval_results_per_group=evaluation_results_per_cl,
            model=model,
        )
    overall_eval = pd.DataFrame.from_dict(overall_eval, orient="index")
    if len(norm_drug_eval_results) > 0:
        overall_eval = _concat_results(norm_drug_eval_results, "drug", overall_eval)
    if len(norm_cl_eval_results) > 0:
        overall_eval = _concat_results(norm_cl_eval_results, "cell_line", overall_eval)

    return (
        overall_eval,
        evaluation_results_per_drug,
        evaluation_results_per_cl,
        true_vs_pred,
        model,
    )


def _concat_results(norm_group_res: dict[str, dict[str, float]], group_by: str, eval_res: pd.DataFrame) -> pd.DataFrame:
    """
    Concatenate the normalized group results to the evaluation results.

    :param norm_group_res: dictionary with the normalized group results, key: model name, value: evaluation results
    :param group_by: either cell line or drug
    :param eval_res: overall dataframe
    :returns: overall dataframe extended by the normalized group results
    """
    norm_group_df = pd.DataFrame.from_dict(norm_group_res, orient="index")
    # append 'group normalized ' to the column names
    norm_group_df.columns = [f"{col}: {group_by} normalized" for col in norm_group_df.columns]
    eval_res = pd.concat([eval_res, norm_group_df], axis=1)
    return eval_res


@pipeline_function
def prep_results(
    eval_results: pd.DataFrame,
    eval_results_per_drug: pd.DataFrame,
    eval_results_per_cell_line: pd.DataFrame,
    t_vs_p: pd.DataFrame,
) -> tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """
    Prepare the results by introducing new columns for algorithm, randomization, setting, split, CV_split.

    :param eval_results: evaluation results
    :param eval_results_per_drug: evaluation results per drug
    :param eval_results_per_cell_line: evaluation results per cell line
    :param t_vs_p: true vs. predicted values
    :returns: the same dataframes with new columns
    """
    # add variables
    # split the index by "_" into: algorithm, randomization, setting, split, CV_split
    new_columns = eval_results.index.str.split("_", expand=True).to_frame()
    new_columns.columns = [
        "algorithm",
        "rand_setting",
        "LPO_LCO_LDO",
        "split",
        "CV_split",
    ]
    new_columns.index = eval_results.index
    eval_results = pd.concat([new_columns.drop("split", axis=1), eval_results], axis=1)
    if eval_results_per_drug is not None:
        eval_results_per_drug[["algorithm", "rand_setting", "LPO_LCO_LDO", "split", "CV_split"]] = (
            eval_results_per_drug["model"].str.split("_", expand=True)
        )
    if eval_results_per_cell_line is not None:
        eval_results_per_cell_line[["algorithm", "rand_setting", "LPO_LCO_LDO", "split", "CV_split"]] = (
            eval_results_per_cell_line["model"].str.split("_", expand=True)
        )
    t_vs_p[["algorithm", "rand_setting", "LPO_LCO_LDO", "split", "CV_split"]] = t_vs_p["model"].str.split(
        "_", expand=True
    )

    return (
        eval_results,
        eval_results_per_drug,
        eval_results_per_cell_line,
        t_vs_p,
    )


def _generate_model_names(test_mode: str, model_name: str, pred_file: pathlib.Path) -> str:
    """
    Generate the model names based on the prediction file.

    :param test_mode: test mode, e.g., LPO
    :param model_name: model name, e.g., SimpleNeuralNetwork
    :param pred_file: file containing the predictions
    :returns: unique name of run = {model_name}_{pred_setting}_{test_mode}_{split}
    :raises ValueError: if the prediction setting is unknown
    """
    file_parts = os.path.basename(pred_file).split("_")
    pred_rand_rob = file_parts[0]
    if pred_rand_rob == "predictions":
        pred_setting = "predictions"
    elif pred_rand_rob == "randomization":
        pred_setting = "randomize-" + "-".join(file_parts[1:-2])
    elif pred_rand_rob == "robustness":
        pred_setting = "-".join(file_parts[:2])
    elif pred_rand_rob == "cross":
        pred_setting = "cross-study-" + file_parts[2]
    else:
        raise ValueError(f"Unknown prediction setting: {pred_rand_rob}")
    split = "_".join(os.path.basename(pred_file).split(".")[0].split("_")[-2:])
    return f"{model_name}_{pred_setting}_{test_mode}_{split}"


def _evaluate_per_group(
    df: pd.DataFrame,
    group_by: str,
    norm_group_eval_results: dict[str, dict[str, float]],
    eval_results_per_group: pd.DataFrame | None,
    model: str,
) -> tuple[dict[str, dict[str, float]], pd.DataFrame]:
    """
    Evaluate the predictions per group.

    :param df: true vs. predicted values
    :param group_by: cell line or drug
    :param norm_group_eval_results: dictionary to store the normalized group evaluation results
    :param eval_results_per_group: evaluation results per group
    :param model: model name
    :returns: dictionary with the normalized group evaluation results and the evaluation results per group
    """
    # calculate the mean of y_true per drug
    print(f"Calculating {group_by}-wise evaluation measures …")
    df[f"mean_y_true_per_{group_by}"] = df.groupby(group_by)["y_true"].transform("mean")
    norm_df = df.copy()
    norm_df["y_true"] = norm_df["y_true"] - norm_df[f"mean_y_true_per_{group_by}"]
    norm_df["y_pred"] = norm_df["y_pred"] - norm_df[f"mean_y_true_per_{group_by}"]
    norm_group_eval_results[model] = evaluate(
        DrugResponseDataset(
            response=norm_df["y_true"].to_numpy(),
            cell_line_ids=norm_df["cell_line"].to_numpy(),
            drug_ids=norm_df["drug"].to_numpy(),
            predictions=norm_df["y_pred"].to_numpy(),
        ),
        list(AVAILABLE_METRICS.keys() - {"MSE", "RMSE", "MAE"}),
    )
    # evaluation per group
    eval_results_per_group = compute_evaluation(df, eval_results_per_group, group_by, model)
    return norm_group_eval_results, eval_results_per_group


def compute_evaluation(df: pd.DataFrame, return_df: pd.DataFrame | None, group_by: str, model: str) -> pd.DataFrame:
    """
    Compute the evaluation metrics per group.

    :param df: true vs. predicted values with mean_y_true_per_{group_by} column
    :param return_df: DataFrame to store the results
    :param group_by: either cell line or drug
    :param model: model name
    :returns: dataframe with the evaluation results per group
    """
    result_per_group = df.groupby(group_by)[["y_true", "cell_line", "drug", "y_pred"]].apply(
        lambda x: evaluate(
            DrugResponseDataset(
                response=x["y_true"].to_numpy(),
                cell_line_ids=x["cell_line"].to_numpy(),
                drug_ids=x["drug"].to_numpy(),
                predictions=x["y_pred"].to_numpy(),
            ),
            list(AVAILABLE_METRICS.keys()),
        )
    )
    groups = result_per_group.index
    result_per_group = pd.json_normalize(result_per_group)
    result_per_group[group_by] = groups
    result_per_group["model"] = model
    if return_df is None:
        return_df = pd.DataFrame(result_per_group)
    else:
        return_df = pd.concat([return_df, result_per_group])
    return return_df


@pipeline_function
def write_results(
    path_out: str,
    eval_results: pd.DataFrame,
    eval_results_per_drug: pd.DataFrame,
    eval_results_per_cl: pd.DataFrame,
    t_vs_p: pd.DataFrame,
) -> None:
    """
    Write the results to csv files.

    :param path_out: path to the output directory, e.g., results/my_run/
    :param eval_results: evaluation results
    :param eval_results_per_drug: evaluation results per drug
    :param eval_results_per_cl: evaluation results per cell line
    :param t_vs_p: true vs. predicted values
    """
    eval_results.to_csv(f"{path_out}evaluation_results.csv", index=True)
    if eval_results_per_drug is not None:
        eval_results_per_drug.to_csv(f"{path_out}evaluation_results_per_drug.csv", index=True)
    if eval_results_per_cl is not None:
        eval_results_per_cl.to_csv(f"{path_out}evaluation_results_per_cl.csv", index=True)
    t_vs_p.to_csv(f"{path_out}true_vs_pred.csv", index=True)


@pipeline_function
def create_index_html(custom_id: str, test_modes: list[str], prefix_results: str) -> None:
    """
    Create the index.html file.

    :param custom_id: custom id for the results, e.g., my_run
    :param test_modes: list of test modes, e.g., ["LPO", "LCO", "LDO"]
    :param prefix_results: path to the results directory, e.g., results/my_run
    """
    # copy images to the results directory
    file_to_copy = [
        "favicon.png",
        "nf-core-drugresponseeval_logo_light.png",
    ]
    for file in file_to_copy:
        file_path = os.path.join(
            str(importlib_resources.files("drevalpy")),
            "visualization",
            "style_utils",
            file,
        )
        shutil.copyfile(file_path, os.path.join(prefix_results, file))

    layout_path = os.path.join(
        str(importlib_resources.files("drevalpy")),
        "visualization",
        "style_utils",
        "index_layout.html",
    )
    idx_html_path = os.path.join(prefix_results, "index.html")
    with open(idx_html_path, "w", encoding="utf-8") as f:
        _parse_layout(f=f, path_to_layout=layout_path)
        f.write('<div class="main">\n')
        f.write('<img src="nf-core-drugresponseeval_logo_light.png" ' 'width="364px" height="100px" alt="Logo">\n')
        f.write(f"<h1>Results for {custom_id}</h1>\n")
        f.write("<h2>Available settings</h2>\n")
        f.write('<div style="display: inline-block;">\n')
        f.write("<p>Click on the images to open the respective report in a new tab.</p>\n")

        test_modes.sort()
        for lpo_lco_ldo in test_modes:
            img_path = os.path.join(
                str(importlib_resources.files("drevalpy")),
                "visualization",
                "style_utils",
                f"{lpo_lco_ldo}.png",
            )
            shutil.copyfile(img_path, os.path.join(prefix_results, f"{lpo_lco_ldo}.png"))
            f.write(
                f'<a href="{lpo_lco_ldo}.html" target="_blank"><img src="{lpo_lco_ldo}.png" '
                f'style="width:300px;height:300px;"></a>\n'
            )
        f.write("</div>\n")
        f.write("</div>\n")
        f.write("</body>\n")
        f.write("</html>\n")


@pipeline_function
def create_html(run_id: str, lpo_lco_ldo: str, files: list, prefix_results: str) -> None:
    """
    Create the html file for the given test mode, e.g., LPO.html.

    :param run_id: custom id for the results, e.g., my_run
    :param lpo_lco_ldo: test mode, e.g., LPO
    :param files: list of files in the results directory
    :param prefix_results: path to the results directory, e.g., results/my_run
    """
    page_layout = os.path.join(
        str(importlib_resources.files("drevalpy")),
        "visualization/style_utils/page_layout.html",
    )
    html_path = os.path.join(prefix_results, f"{lpo_lco_ldo}.html")

    with open(html_path, "w", encoding="utf-8") as f:
        _parse_layout(f=f, path_to_layout=page_layout)
        f.write(f"<h1>Results for {run_id}: {lpo_lco_ldo}</h1>\n")

        # Critical difference plot
        f = CriticalDifferencePlot.write_to_html(lpo_lco_ldo=lpo_lco_ldo, f=f)

        # Violin plots
        f = VioHeat.write_to_html(lpo_lco_ldo=lpo_lco_ldo, f=f, files=files, plot="Violin")

        # Heatmaps
        f = VioHeat.write_to_html(lpo_lco_ldo=lpo_lco_ldo, f=f, files=files, plot="Heatmap")

        # Regression plots
        f = RegressionSliderPlot.write_to_html(lpo_lco_ldo=lpo_lco_ldo, f=f, files=files)

        # Correlation comparison: Drug
        f = CorrelationComparisonScatter.write_to_html(lpo_lco_ldo=lpo_lco_ldo, f=f, files=files)

        # Evaluation results tables
        f = HTMLTable.write_to_html(lpo_lco_ldo=lpo_lco_ldo, f=f, files=files, prefix=prefix_results)

        f.write("</div>\n")
        f.write("</body>\n")
        f.write("</html>\n")



"""Plots a heatmap of the evaluation metrics."""

import numpy as np
import pandas as pd
import plotly.graph_objects as go
from plotly.subplots import make_subplots

from ..pipeline_function import pipeline_function
from .vioheat import VioHeat


class Heatmap(VioHeat):
    """Plots a heatmap of the evaluation metrics."""

    @pipeline_function
    def __init__(self, df: pd.DataFrame, normalized_metrics=False, whole_name=False):
        """
        Initialize the Heatmap class.

        :param df: either containing all predictions for all algorithms or all tests for one algorithm (including
            robustness, randomization, … tests then)
        :param normalized_metrics: whether the metrics are normalized
        :param whole_name: whether the whole name should be displayed
        """
        super().__init__(df, normalized_metrics, whole_name)
        self.df = self.df[[col for col in self.df.columns if col in self.all_metrics]]
        if self.normalized_metrics:
            titles = [
                "Standard Errors over CV folds",
                "Mean R^2: normalized",
                "Mean Correlations: normalized",
            ]
            nr_subplots = 3
            self.plot_settings = ["standard_errors", "r2", "correlations"]
            self.fig = make_subplots(
                rows=nr_subplots,
                cols=1,
                subplot_titles=tuple(titles),
                vertical_spacing=0.25,
            )
        else:
            titles = [
                "Standard Errors over CV folds",
                "Mean R^2",
                "Mean Correlations",
                "Mean Errors",
            ]
            nr_subplots = 4
            self.plot_settings = [
                "standard_errors",
                "r2",
                "correlations",
                "errors",
            ]
            self.fig = make_subplots(
                rows=nr_subplots,
                cols=1,
                subplot_titles=tuple(titles),
                vertical_spacing=0.1,
            )

    @pipeline_function
    def draw_and_save(self, out_prefix: str, out_suffix: str) -> None:
        """
        Draw the heatmap and save it to a file.

        :param out_prefix: e.g., results/my_run/heatmaps/
        :param out_suffix: e.g., algorithms_normalized
        """
        self._draw()
        path_out = f"{out_prefix}heatmap_{out_suffix}.html"
        self.fig.write_html(path_out)

    def _draw(self) -> None:
        """Draw the heatmap."""
        print("Drawing heatmaps ...")
        for plot_setting in self.plot_settings:
            self._draw_subplots(plot_setting)
        self.fig.update_layout(
            height=1000,
            width=1100,
            title_text="Heatmap of the evaluation metrics",
        )
        self.fig.update_traces(showscale=False)

    def _draw_subplots(self, plot_setting: str) -> None:
        """
        Draw the subplots of the heatmap.

        :param plot_setting: Either "standard_errors", "r2", "correlations", or "errors"
        :raises ValueError: If an unknown plot setting is given
        """
        idx_split = self.df.index.to_series().str.split("_")
        setting = idx_split.str[0:3].str.join("_")
        if plot_setting == "standard_errors":
            dt = self.df.groupby(setting).apply(lambda x: self._calc_summary_metric(x=x, std_error=True))
            row_idx = 1
            colorscale = "Pinkyl"
        elif plot_setting == "r2":
            r2_columns = [col for col in self.df.columns if "R^2" in col]
            dt = self.df[r2_columns]
            dt = dt.groupby(setting).apply(lambda x: self._calc_summary_metric(x=x, std_error=False))
            dt = dt.sort_values(by=r2_columns[0], ascending=True)
            row_idx = 2
            colorscale = "Blues"
        elif plot_setting == "correlations":
            corr_columns = [
                col
                for col in self.df.columns
                if "Pearson" in col or "Spearman" in col or "Kendall" in col or "Partial_Correlation" in col
            ]
            corr_columns.sort()
            dt = self.df[corr_columns]
            dt = dt.groupby(setting).apply(lambda x: self._calc_summary_metric(x=x, std_error=False))
            dt = dt.sort_values(by=corr_columns[0], ascending=True)
            row_idx = 3
            colorscale = "Viridis"
        elif plot_setting == "errors":
            dt = self.df[["MSE", "RMSE", "MAE"]]
            dt = dt.groupby(setting).apply(lambda x: self._calc_summary_metric(x=x, std_error=False))
            dt = dt.sort_values(by="MSE", ascending=False)
            row_idx = 4
            colorscale = "hot"
        else:
            raise ValueError("Unknown plot setting")
        if self.whole_name:
            labels = [i.replace("_", " ") for i in list(dt.index)]
        else:
            labels = [i.split("_")[0] for i in list(dt.index)]
        self.fig = self.fig.add_trace(
            go.Heatmap(
                z=dt.values,
                x=dt.columns,
                y=labels,
                colorscale=colorscale,
                texttemplate="%{z:.2f}",
            ),
            row=row_idx,
            col=1,
        )

    @staticmethod
    def _calc_summary_metric(x: pd.DataFrame, std_error: bool = False):
        """
        Calculate the mean or standard error of the metrics.

        :param x: DataFrame containing the metrics
        :param std_error: whether to calculate the standard error or the mean
        :returns: Series containing the mean or standard error of the metrics
        """
        # make empty results series
        results = pd.Series(index=x.columns)
        # iterate over columns
        for col in x.columns:
            if np.count_nonzero(np.isnan(x[col])) == len(x[col]):
                results[col] = np.nan
            elif std_error:
                # calculate standard error
                results[col] = np.nanstd(x[col]) / np.sqrt(x.shape[0])
            else:
                # calculate mean
                results[col] = np.nanmean(x[col])
        return results



"""Contains functions to load the GDSC1, GDSC2, CCLE, and Toy datasets."""

import os
from pathlib import Path
from typing import Callable

import pandas as pd

from ..pipeline_function import pipeline_function
from .curvecurator import fit_curves
from .dataset import DrugResponseDataset
from .utils import download_dataset


def load_gdsc1(
    path_data: str = "data",
    measure: str = "LN_IC50",
    file_name: str = "response_GDSC1.csv",
    dataset_name: str = "GDSC1",
) -> DrugResponseDataset:
    """
    Loads the GDSC1 dataset.

    :param path_data: Path to the dataset.
    :param file_name: File name of the dataset.
    :param measure: The name of the column containing the measure to predict, default = "LN_IC50"

    :param dataset_name: Name of the dataset.
    :return: DrugResponseDataset containing response, cell line IDs, and drug IDs.
    """
    path = os.path.join(path_data, dataset_name, file_name)
    if not os.path.exists(path):
        download_dataset(dataset_name, path_data, redownload=True)

    response_data = pd.read_csv(path)
    response_data["DRUG_NAME"] = response_data["DRUG_NAME"].str.replace(",", "")

    return DrugResponseDataset(
        response=response_data[measure].values,
        cell_line_ids=response_data["CELL_LINE_NAME"].values,
        drug_ids=response_data["DRUG_NAME"].values,
        dataset_name=dataset_name,
    )


def load_gdsc2(path_data: str = "data", measure: str = "LN_IC50", file_name: str = "response_GDSC2.csv"):
    """
    Loads the GDSC2 dataset.

    :param path_data: Path to the dataset.
    :param file_name: File name of the dataset.
    :param measure: The name of the column containing the measure to predict, default = "LN_IC50"

    :return: DrugResponseDataset containing response, cell line IDs, and drug IDs.
    """
    return load_gdsc1(path_data=path_data, measure=measure, file_name=file_name, dataset_name="GDSC2")


def load_ccle(
    path_data: str = "data", measure: str = "LN_IC50", file_name: str = "response_CCLE.csv"
) -> DrugResponseDataset:
    """
    Loads the CCLE dataset.

    :param path_data: Path to the dataset.
    :param file_name: File name of the dataset.
    :param measure: The name of the column containing the measure to predict, default = "LN_IC50"

    :return: DrugResponseDataset containing response, cell line IDs, and drug IDs.
    """
    dataset_name = "CCLE"
    path = os.path.join(path_data, dataset_name, file_name)
    if not os.path.exists(path):
        download_dataset(dataset_name, path_data, redownload=True)

    response_data = pd.read_csv(path)
    response_data["DRUG_NAME"] = response_data["DRUG_NAME"].str.replace(",", "")

    return DrugResponseDataset(
        response=response_data[measure].values,
        cell_line_ids=response_data["CELL_LINE_NAME"].values,
        drug_ids=response_data["DRUG_NAME"].values,
        dataset_name=dataset_name,
    )


def load_toy(path_data: str = "data", measure: str = "response") -> DrugResponseDataset:
    """
    Loads small Toy dataset, subsampled from GDSC1.

    :param path_data: Path to the dataset.
    :param measure: The name of the column containing the measure to predict, default = "response"

    :return: DrugResponseDataset containing response, cell line IDs, and drug IDs.
    """
    dataset_name = "Toy_Data"
    measure = "response"  # overwrite this explicitly to avoid problems, should be changed in the future
    path = os.path.join(path_data, dataset_name, "toy_data.csv")
    if not os.path.exists(path):
        download_dataset(dataset_name, path_data, redownload=True)
    response_data = pd.read_csv(path)

    return DrugResponseDataset(
        response=response_data[measure].values,
        cell_line_ids=response_data["cell_line_id"].values,
        drug_ids=response_data["drug_id"].values,
        dataset_name=dataset_name,
    )


def load_custom(path_data: str | Path, measure: str = "response") -> DrugResponseDataset:
    """
    Load custom dataset.

    :param path_data: Path to location of custom dataset
    :param measure: The name of the column containing the measure to predict, default = "response"

    :return: DrugResponseDataset containing response, cell line IDs, and drug IDs
    """
    return DrugResponseDataset.from_csv(path_data, measure=measure)


AVAILABLE_DATASETS: dict[str, Callable] = {
    "GDSC1": load_gdsc1,
    "GDSC2": load_gdsc2,
    "CCLE": load_ccle,
    "Toy_Data": load_toy,
}


@pipeline_function
def load_dataset(
    dataset_name: str, path_data: str = "data", measure: str = "response", curve_curator: bool = False, cores: int = 1
) -> DrugResponseDataset:
    """
    Load a dataset based on the dataset name.

    :param dataset_name: The name of the dataset to load. Can be one of ('GDSC1', 'GDSC2', 'CCLE', or 'Toy_Data')
        to download provided datasets, or any other name to allow for custom datasets.
    :param path_data: The parent path in which custom or downloaded datasets should be located, or in which raw
        viability data is to be found for fitting with CurveCurator (see param curve_curator for details).
        The location of the datasets are resolved by <path_data>/<dataset_name>/<dataset_name>.csv.
    :param measure: The name of the column containing the measure to predict, default = "response".
        If curve_curator is True, this measure is appended with "_curvecurator", e.g. "response_curvecurator" to
        distinguish between measures provided by the original source of a dataset, or the measures fit by
        CurveCurator.
    :param curve_curator: If True, the measure is appended with "_curvecurator".
        If a custom dataset_name was provided, this will invoke the fitting procedure of raw viability data,
        which is expected to exist at <path_data>/<dataset_name>/<dataset_name>_raw.csv. The fitted dataset will
        be stored in the same folder, in a file called <dataset_name>.csv
    :param cores: Number of cores to use for CurveCurator fitting. Only used when curve_curator is True, default = 1
    :return: A DrugResponseDataset containing response, cell line IDs, drug IDs, and dataset name.
    :raises FileNotFoundError: If the custom dataset or raw viability data could not be found at the given path.
    """
    if curve_curator:
        measure += "_curvecurator"
        input_file = Path(path_data) / dataset_name / f"{dataset_name}_raw.csv"
    else:
        input_file = Path(path_data) / dataset_name / f"{dataset_name}.csv"

    if dataset_name in AVAILABLE_DATASETS:
        return AVAILABLE_DATASETS[dataset_name](path_data)

    if input_file.is_file():
        if curve_curator:
            fit_curves(
                input_file=input_file,
                output_dir=input_file.parent,
                dataset_name=dataset_name,
                cores=cores,
            )
        return load_custom(Path(path_data) / dataset_name / f"{dataset_name}.csv", measure=measure)
    raise FileNotFoundError(f"Custom dataset does not exist at given path: {input_file}")



"""Utility functions for the SuperFELTR model."""

import os
import secrets

import numpy as np
import pytorch_lightning as pl
import torch
from pytorch_lightning.callbacks import EarlyStopping, TQDMProgressBar
from torch import nn

from ...datasets.dataset import DrugResponseDataset, FeatureDataset
from ..MOLIR.utils import create_dataset_and_loaders, generate_triplets_indices


class SuperFELTEncoder(pl.LightningModule):
    """
    SuperFELT encoder definition for a single omic type, i.e., gene expression, mutation, or copy number variation.

    Very similar to MOLIEncoder, but with BatchNorm1d before ReLU.
    """

    def __init__(
        self, input_size: int, hpams: dict[str, int | float | dict], omic_type: str, ranges: tuple[float, float]
    ) -> None:
        """
        Initializes the SuperFELTEncoder.

        Save_hyperparameters is turned on to facilitate loading the model from a checkpoint.
        :param input_size: determined by the variance threshold feature selection
        :param hpams: hyperparameters for the model
        :param omic_type: gene expression, mutation, or copy number variation
        :param ranges: positive and negative ranges for the triplet loss
        :raises ValueError: if the hyperparameters are not of the correct type
        """
        super().__init__()
        self.save_hyperparameters()
        if (
            not isinstance(hpams["dropout_rate"], float)
            or not isinstance(hpams["margin"], float)
            or not isinstance(hpams["learning_rate"], float)
            or not isinstance(hpams["weight_decay"], float)
        ):
            raise ValueError("dropout_rate, margin, learning_rate, and weight_decay must be floats!")

        self.omic_type = omic_type
        output_size = self._get_output_size(hpams)

        # only change vs MOLIEncoder: BatchNorm1d before ReLU
        self.encode = nn.Sequential(
            nn.Linear(input_size, output_size),
            nn.BatchNorm1d(output_size),
            nn.ReLU(),
            nn.Dropout(hpams["dropout_rate"]),
        )
        self.lr = hpams["learning_rate"]
        self.weight_decay = hpams["weight_decay"]
        self.triplet_loss = nn.TripletMarginLoss(margin=hpams["margin"], p=2)
        self.positive_range, self.negative_range = ranges

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the SuperFELTEncoder.

        :param x: input tensor
        :returns: encoded tensor
        """
        return self.encode(x)

    def configure_optimizers(self) -> torch.optim.Optimizer:
        """
        Override the configure_optimizers method to use the Adam optimizer.

        :returns: Adam optimizer
        """
        optimizer = torch.optim.Adam(self.parameters(), lr=self.lr, weight_decay=self.weight_decay)
        return optimizer

    def _get_output_size(self, hpams: dict[str, int | float | dict]) -> int:
        """
        Get the output size of the encoder based on the omic type from the hyperparameters.

        :param hpams: hyperparameters for the model
        :returns: output size of the encoder
        :raises ValueError: if the output sizes are not of the correct type
        """
        if (
            not isinstance(hpams["out_dim_expr_encoder"], int)
            or not isinstance(hpams["out_dim_mutation_encoder"], int)
            or not isinstance(hpams["out_dim_cnv_encoder"], int)
        ):
            raise ValueError("out_dim_expr_encoder, out_dim_mutation_encoder, and out_dim_cnv_encoder must be ints!")

        output_sizes = {
            "expression": hpams["out_dim_expr_encoder"],
            "mutation": hpams["out_dim_mutation_encoder"],
            "copy_number_variation_gistic": hpams["out_dim_cnv_encoder"],
        }
        output_size = output_sizes[self.omic_type]
        return output_size

    def _get_omic_data(self, data_expr: torch.Tensor, data_mut: torch.Tensor, data_cnv: torch.Tensor) -> torch.Tensor:
        """
        Get the omic data based on the omic type.

        :param data_expr: expression data
        :param data_mut: mutation data
        :param data_cnv: copy number variation data
        :returns: the omic data
        :raises ValueError: if the omic type is not recognized
        """
        if self.omic_type == "expression":
            data = data_expr
        elif self.omic_type == "mutation":
            data = data_mut
        elif self.omic_type == "copy_number_variation_gistic":
            data = data_cnv
        else:
            raise ValueError(f"omic_type {self.omic_type} not recognized.")
        return data

    def _compute_loss(self, encoded: torch.Tensor, response: torch.Tensor) -> torch.Tensor:
        """
        Computes the triplet loss.

        :param encoded: encoded data
        :param response: response data
        :returns: triplet loss
        """
        positive_indices, negative_indices = generate_triplets_indices(
            response.cpu().detach().numpy(), self.positive_range, self.negative_range
        )
        triplet_loss = self.triplet_loss(encoded, encoded[positive_indices], encoded[negative_indices])
        return triplet_loss

    def training_step(self, batch: list[torch.Tensor], batch_idx: int) -> torch.Tensor:
        """
        Override the training_step method to compute the triplet loss.

        :param batch: batch containing the omic data and response
        :param batch_idx: index of the batch
        :returns: triplet loss
        """
        data_expr, data_mut, data_cnv, response = batch
        data = self._get_omic_data(data_expr, data_mut, data_cnv)
        encoded = self.encode(data)
        triplet_loss = self._compute_loss(encoded, response)
        self.log("train_loss", triplet_loss, on_step=False, on_epoch=True, prog_bar=True)
        return triplet_loss

    def validation_step(self, batch: list[torch.Tensor], batch_idx: int) -> torch.Tensor:
        """
        Override the validation_step method to compute the triplet loss.

        :param batch: batch containing the omic data and response
        :param batch_idx: index of the batch
        :returns: triplet loss
        """
        data_expr, data_mut, data_cnv, response = batch
        data = self._get_omic_data(data_expr, data_mut, data_cnv)
        encoded = self.encode(data)
        triplet_loss = self._compute_loss(encoded, response)
        self.log("val_loss", triplet_loss, on_step=False, on_epoch=True, prog_bar=True)
        return triplet_loss


class SuperFELTRegressor(pl.LightningModule):
    """
    SuperFELT regressor definition.

    Very similar to SuperFELT classifier, but with a regression loss and without the last sigmoid layer.
    """

    def __init__(
        self,
        input_size: int,
        hpams: dict[str, int | float | dict],
        encoders: tuple[SuperFELTEncoder, SuperFELTEncoder, SuperFELTEncoder],
    ) -> None:
        """
        Initializes the SuperFELTRegressor.

        The encoders are put in eval mode because they were fitted before.

        :param input_size: depends on the output of the encoders
        :param hpams: hyperparameters for the model
        :param encoders: the fitted encoders for the gene expression, mutation, and copy number variation data
        :raises ValueError: if the hyperparameters are not of the correct type
        """
        super().__init__()
        if (
            not isinstance(hpams["learning_rate"], float)
            or not isinstance(hpams["weight_decay"], float)
            or not isinstance(hpams["dropout_rate"], float)
        ):
            raise ValueError("learning_rate, weight_decay and dropout_rate must be floats!")

        self.regressor = nn.Sequential(nn.Linear(input_size, 1), nn.Dropout(hpams["dropout_rate"]))
        self.lr = float(hpams["learning_rate"])
        self.weight_decay = float(hpams["weight_decay"])
        self.encoders = encoders
        # put the encoders in eval mode
        for encoder in self.encoders:
            encoder.eval()
        self.regression_loss = nn.MSELoss()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the SuperFELTRegressor.

        :param x: input tensor
        :returns: predicted response
        """
        return self.regressor(x)

    def predict(self, data_expr: np.ndarray, data_mut: np.ndarray, data_cnv: np.ndarray) -> np.ndarray:
        """
        Predicts the response for the given input.

        :param data_expr: expression data
        :param data_mut: mutation data
        :param data_cnv: copy number variation data
        :returns: predicted response
        """
        data_expr_tensor, data_mut_tensor, data_cnv_tensor = map(
            lambda data: torch.from_numpy(data).float().to(self.device), [data_expr, data_mut, data_cnv]
        )
        self.eval()
        with torch.no_grad():
            encoded = self._encode_and_concatenate(data_expr_tensor, data_mut_tensor, data_cnv_tensor)
            preds = self.regressor(encoded)
        return preds.squeeze().cpu().detach().numpy()

    def configure_optimizers(self) -> torch.optim.Optimizer:
        """
        Override the configure_optimizers method to use the Adagrad optimizer.

        :returns: Adagrad optimizer
        """
        return torch.optim.Adagrad(self.parameters(), lr=self.lr, weight_decay=self.weight_decay)

    def _encode_and_concatenate(
        self, data_expr: torch.Tensor, data_mut: torch.Tensor, data_cnv: torch.Tensor
    ) -> torch.Tensor:
        """
        Encodes the omic data and concatenates the encoded tensors.

        :param data_expr: expression data
        :param data_mut: mutation data
        :param data_cnv: copy number variation data
        :returns: concatenated encoded tensor
        """
        encoded_expr = self.encoders[0].encode(data_expr)
        encoded_mut = self.encoders[1].encode(data_mut)
        encoded_cnv = self.encoders[2].encode(data_cnv)
        return torch.cat((encoded_expr, encoded_mut, encoded_cnv), dim=1)

    def training_step(self, batch: list[torch.Tensor], batch_idx: int) -> torch.Tensor:
        """
        Override the training_step method to compute the regression loss.

        :param batch: batch containing the omic data and response
        :param batch_idx: index of the batch
        :returns: regression loss
        """
        data_expr, data_mut, data_cnv, response = batch
        encoded = self._encode_and_concatenate(data_expr, data_mut, data_cnv)
        pred = self.regressor(encoded)
        loss = self.regression_loss(pred.squeeze(), response)
        self.log("train_loss", loss, on_step=False, on_epoch=True, prog_bar=True)
        return loss

    def validation_step(self, batch: list[torch.Tensor], batch_idx: int) -> torch.Tensor:
        """
        Override the validation_step method to compute the regression loss.

        :param batch: batch containing the omic data and response
        :param batch_idx: index of the batch
        :returns: regression loss
        """
        data_expr, data_mut, data_cnv, response = batch
        encoded = self._encode_and_concatenate(data_expr, data_mut, data_cnv)
        pred = self.regressor(encoded)
        loss = self.regression_loss(pred.squeeze(), response)
        self.log("val_loss", loss, on_step=False, on_epoch=True, prog_bar=True)
        return loss


def train_superfeltr_model(
    model: SuperFELTEncoder | SuperFELTRegressor,
    hpams: dict[str, int | float | dict],
    output_train: DrugResponseDataset,
    cell_line_input: FeatureDataset,
    output_earlystopping: DrugResponseDataset | None = None,
    patience: int = 5,
    model_checkpoint_dir: str = "superfeltr_checkpoints",
) -> pl.callbacks.ModelCheckpoint:
    """
    Trains one encoder or the regressor.

    First, the dataset and loaders are created. Then, the model is trained with the Lightning trainer.
    :param model: either one of the encoders or the regressor
    :param hpams: hyperparameters for the model
    :param output_train: response data for training
    :param cell_line_input: cell line omics features
    :param output_earlystopping: response data for early stopping
    :param patience: for early stopping, defaults to 5
    :param model_checkpoint_dir: directory to save the model checkpoints
    :returns: checkpoint callback with the best model
    :raises ValueError: if the epochs and mini_batch are not integers
    """
    if not isinstance(hpams["epochs"], int) or not isinstance(hpams["mini_batch"], int):
        raise ValueError("epochs and mini_batch must be integers!")

    train_loader, val_loader = create_dataset_and_loaders(
        batch_size=hpams["mini_batch"],
        output_train=output_train,
        cell_line_input=cell_line_input,
        output_earlystopping=output_earlystopping,
    )
    monitor = "train_loss" if (val_loader is None) else "val_loss"
    early_stop_callback = EarlyStopping(monitor=monitor, mode="min", patience=patience)
    name = "version-" + "".join(
        [secrets.choice("0123456789abcdef") for _ in range(20)]
    )  # preventing conflicts of filenames
    checkpoint_callback = pl.callbacks.ModelCheckpoint(
        dirpath=os.path.join(model_checkpoint_dir, name),
        monitor=monitor,
        mode="min",
        save_top_k=1,
    )
    # Initialize the Lightning trainer
    trainer = pl.Trainer(
        max_epochs=hpams["epochs"],
        callbacks=[
            early_stop_callback,
            checkpoint_callback,
            TQDMProgressBar(refresh_rate=0),
        ],
    )
    if val_loader is None:
        trainer.fit(model, train_loader)
    else:
        trainer.fit(model, train_loader, val_loader)
    return checkpoint_callback



"""
Contains the SingleDrugRandomForest class.

It is a RandomForest model that uses only gene expression dataset for drug response prediction and trains one model
per drug.
"""

import numpy as np

from ...datasets.dataset import DrugResponseDataset, FeatureDataset
from .sklearn_models import RandomForest


class SingleDrugRandomForest(RandomForest):
    """SingleDrugRandomForest class."""

    is_single_drug_model = True
    drug_views = []
    early_stopping = False

    @classmethod
    def get_model_name(cls) -> str:
        """
        Returns the model name.

        :returns: SingleDrugRandomForest
        """
        return "SingleDrugRandomForest"

    def train(
        self,
        output: DrugResponseDataset,
        cell_line_input: FeatureDataset,
        drug_input: FeatureDataset | None = None,
        output_earlystopping: DrugResponseDataset | None = None,
        model_checkpoint_dir: str = "checkpoints",
    ) -> None:
        """
        Trains the model; the number of features is the number of fingerprints.

        :param output: training dataset containing the response output
        :param cell_line_input: training dataset containing gene expression data
        :param drug_input: not needed
        :param output_earlystopping: not needed
        :param model_checkpoint_dir: not needed as checkpoints are not saved
        :raises ValueError: if drug_input is not None
        """
        if drug_input is not None:
            raise ValueError("SingleDrugRandomForest does not support drug_input!")

        if len(output) > 0:
            x = self.get_concatenated_features(
                cell_line_view="gene_expression",
                drug_view=None,
                cell_line_ids_output=output.cell_line_ids,
                drug_ids_output=output.drug_ids,
                cell_line_input=cell_line_input,
                drug_input=None,
            )
            self.model.fit(x, output.response)
        else:
            print("No training data provided, will predict NA.")
            self.model = None

    def predict(
        self,
        cell_line_ids: np.ndarray,
        drug_ids: np.ndarray,
        cell_line_input: FeatureDataset,
        drug_input: FeatureDataset | None = None,
    ) -> np.ndarray:
        """
        Predicts the drug response for the given cell lines.

        :param cell_line_ids: cell line ids
        :param drug_ids: drug ids, not needed here
        :param cell_line_input: cell line input
        :param drug_input: drug input, not needed here
        :returns: predicted drug response
        :raises ValueError: if drug_input is not None
        """
        if drug_input is not None:
            raise ValueError("drug_input is not needed.")

        if self.model is None:
            print("No training data was available, predicting NA.")
            return np.array([np.nan] * len(cell_line_ids))
        x = self.get_concatenated_features(
            cell_line_view="gene_expression",
            drug_view=None,
            cell_line_ids_output=cell_line_ids,
            drug_ids_output=drug_ids,
            cell_line_input=cell_line_input,
            drug_input=None,
        )
        return self.model.predict(x)



"""Module for generating regression plots with a slider for Pearson correlation coefficient."""

from io import TextIOWrapper

import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from scipy.stats import pearsonr

from ..pipeline_function import pipeline_function
from .outplot import OutPlot


class RegressionSliderPlot(OutPlot):
    """Generates regression plots with a slider for the Pearson correlation coefficient."""

    @pipeline_function
    def __init__(
        self,
        df: pd.DataFrame,
        lpo_lco_ldo: str,
        model: str,
        group_by: str = "drug",
        normalize=False,
    ):
        """
        Initialize the RegressionSliderPlot class.

        :param df: true vs. predicted values
        :param lpo_lco_ldo: setting, e.g., LPO
        :param model: model name
        :param group_by: either "drug" or "cell_line"
        :param normalize: whether to normalize the true and predicted values by the mean of the group
        """
        self.df = df[(df["LPO_LCO_LDO"] == lpo_lco_ldo) & (df["rand_setting"] == "predictions")]
        self.df = self.df[(self.df["algorithm"] == model)]
        self.group_by = group_by
        self.normalize = normalize
        self.fig = go.Figure()
        self.model = model

        if self.normalize:
            if self.group_by == "cell_line":
                self.df.loc[:, "y_true"] = self.df["y_true"] - self.df["mean_y_true_per_drug"]
                self.df.loc[:, "y_pred"] = self.df["y_pred"] - self.df["mean_y_true_per_drug"]
            else:
                self.df.loc[:, "y_true"] = self.df["y_true"] - self.df["mean_y_true_per_cell_line"]
                self.df.loc[:, "y_pred"] = self.df["y_pred"] - self.df["mean_y_true_per_cell_line"]

    @pipeline_function
    def draw_and_save(self, out_prefix: str, out_suffix: str) -> None:
        """
        Draw the regression plot and save it to a file.

        :param out_prefix: e.g., results/my_run/regression_plots/
        :param out_suffix: e.g., LPO_drug_SimpleNeuralNetwork
        """
        self._draw()
        self.fig.write_html(f"{out_prefix}regression_lines_{out_suffix}.html")

    def _draw(self):
        """Draw the regression plot."""
        print(f"Generating regression plots for {self.group_by}, normalize={self.normalize}...")
        self.df = self.df.groupby(self.group_by).filter(lambda x: len(x) > 1)
        pccs = self.df.groupby(self.group_by).apply(lambda x: pearsonr(x["y_true"], x["y_pred"])[0])
        pccs = pccs.reset_index()
        pccs.columns = [self.group_by, "pcc"]
        self.df = self.df.merge(pccs, on=self.group_by)
        self._render_plot()

    @staticmethod
    def write_to_html(lpo_lco_ldo: str, f: TextIOWrapper, *args, **kwargs) -> TextIOWrapper:
        """
        Write the plot to the final report file.

        :param lpo_lco_ldo: setting, e.g., LPO
        :param f: final report file
        :param args: additional arguments
        :param kwargs: additional keyword arguments, in this case all files
        :return: the final report file
        """
        files: list[str] = kwargs.get("files", [])
        f.write('<h2 id="regression_plots">Regression plots</h2>\n')
        f.write("<ul>\n")
        regr_files = [f for f in files if lpo_lco_ldo in f and f.startswith("regression_lines")]
        regr_files.sort()
        for regr_file in regr_files:
            f.write(f'<li><a href="regression_plots/{regr_file}" target="_blank">{regr_file}</a></li>\n')
        f.write("</ul>\n")
        return f

    def _render_plot(self):
        """Render the regression plot."""
        # sort df by group name
        df = self.df.sort_values(self.group_by)
        setting_title = self.model + " " + df["LPO_LCO_LDO"].unique()[0]
        if self.normalize:
            if self.group_by == "cell_line":
                setting_title += ", normalized by drug mean"
                hover_data = [
                    "pcc",
                    "cell_line",
                    "drug",
                    "mean_y_true_per_drug",
                    "algorithm",
                ]
            else:
                setting_title += ", normalized by cell line mean"
                hover_data = [
                    "pcc",
                    "cell_line",
                    "drug",
                    "mean_y_true_per_cell_line",
                    "algorithm",
                ]

        else:
            hover_data = ["pcc", "cell_line", "drug", "algorithm"]
        self.fig = px.scatter(
            df,
            x="y_true",
            y="y_pred",
            color=self.group_by,
            trendline="ols",
            hover_name=self.group_by,
            hover_data=hover_data,
            title=f"{setting_title}: Regression plot",
        )

        min_val = np.min([np.min(df["y_true"]), np.min(df["y_pred"])])
        max_val = np.max([np.max(df["y_true"]), np.max(df["y_pred"])])
        self.fig.update_xaxes(range=[min_val, max_val])
        self.fig.update_yaxes(range=[min_val, max_val])
        self._make_slider(setting_title)

    def _make_slider(self, setting_title: str) -> None:
        """
        Make a slider for the Pearson correlation coefficient.

        :param setting_title: title of the plot
        """
        n_ticks = 21
        steps = []
        # take the range from pcc (-1 - 1) and divide it into n_ticks-1 equal parts
        pcc_parts = np.linspace(-1, 1, n_ticks)
        for i in range(n_ticks):
            # from the fig data, get the hover data and check if it is greater than the pcc_parts[i]
            # only iterate over even numbers because there are scatter points and the ols line for each group
            pccs = [0 for _ in range(0, len(self.fig.data))]
            for j in range(0, len(self.fig.data)):
                if j % 2 == 0:
                    pccs[j] = self.fig.data[j].customdata[0, 0]
                else:
                    pccs[j] = self.fig.data[j - 1].customdata[0, 0]
            if i == n_ticks - 1:
                # last step
                visible_traces = pccs >= pcc_parts[i]
                title = (
                    f"{setting_title}: Slider for PCCs >= {str(round(pcc_parts[i], 1))} (step {str(i + 1)} "
                    f"of {str(n_ticks)})"
                )
            else:
                # get traces between pcc_parts[i] and pcc_parts[i+1]
                visible_traces_gt = pccs >= pcc_parts[i]
                visible_traces_lt = pccs < pcc_parts[i + 1]
                visible_traces = visible_traces_gt & visible_traces_lt
                title = (
                    f"{setting_title}: Slider for PCCs between {str(round(pcc_parts[i], 1))} "
                    f"and {str(round(pcc_parts[i + 1], 1))} (step {str(i + 1)} of {str(n_ticks)})"
                )
            step = dict(
                method="update",
                args=[{"visible": visible_traces}, {"title": title}],
                label=str(round(pcc_parts[i], 1)),
            )
            steps.append(step)

        sliders = [
            dict(
                active=0,
                currentvalue={"prefix": "Pearson correlation coefficient="},
                pad={"t": 50},
                steps=steps,
            )
        ]

        self.fig.update_layout(
            sliders=sliders,
            legend=dict(yanchor="top", y=1.0, xanchor="left", x=1.05),
        )



"""Renders the evaluation results as HTML tables."""

import os
from io import TextIOWrapper

import pandas as pd

from ..pipeline_function import pipeline_function
from .outplot import OutPlot


class HTMLTable(OutPlot):
    """Renders the evaluation results as HTML tables."""

    @pipeline_function
    def __init__(self, df: pd.DataFrame, group_by: str):
        """
        Initialize the HTMLTable class.

        :param df: either all results of a setting or results evaluated by group (cell line, drug) for a setting
        :param group_by: all or the group by which the results are evaluated
        """
        self.df = df
        self.group_by = group_by

    @pipeline_function
    def draw_and_save(self, out_prefix: str, out_suffix: str) -> None:
        """
        Draw the table and save it to a file.

        :param out_prefix: e.g., results/my_run/html_tables/
        :param out_suffix: e.g., LPO, LPO_drug
        """
        self._draw()
        path_out = f"{out_prefix}table_{out_suffix}.html"
        self.df.to_html(path_out, index=False)

    def _draw(self) -> None:
        """Draw the table."""
        selected_columns = [
            "algorithm",
            "rand_setting",
            "CV_split",
            "MSE",
            "R^2",
            "Pearson",
            "RMSE",
            "MAE",
            "Spearman",
            "Kendall",
            "Partial_Correlation",
            "LPO_LCO_LDO",
        ]
        if self.group_by == "drug":
            selected_columns = ["drug"] + selected_columns
        elif self.group_by == "cell_line":
            selected_columns = ["cell_line"] + selected_columns
        else:
            selected_columns = [
                "algorithm",
                "rand_setting",
                "CV_split",
                "MSE",
                "R^2",
                "Pearson",
                "R^2: drug normalized",
                "Pearson: drug normalized",
                "R^2: cell_line normalized",
                "Pearson: cell_line normalized",
                "RMSE",
                "MAE",
                "Spearman",
                "Kendall",
                "Partial_Correlation",
                "Spearman: drug normalized",
                "Kendall: drug normalized",
                "Partial_Correlation: drug normalized",
                "Spearman: cell_line normalized",
                "Kendall: cell_line normalized",
                "Partial_Correlation: cell_line normalized",
                "LPO_LCO_LDO",
            ]
            # only take the columns that occur
            selected_columns = [col for col in selected_columns if col in self.df.columns]
        # reorder columns
        self.df = self.df[selected_columns]

    @staticmethod
    def write_to_html(lpo_lco_ldo: str, f: TextIOWrapper, prefix: str = "", *args, **kwargs) -> TextIOWrapper:
        """
        Write the evaluation results into the report HTML file.

        :param lpo_lco_ldo: setting, e.g., LPO
        :param f: report file
        :param prefix: e.g., results/my_run
        :param args: additional arguments
        :param kwargs: additional keyword arguments
        :return: the report file
        """
        files: list[str] = kwargs.get("files", [])
        if prefix != "":
            prefix = os.path.join(prefix, "html_tables")
        f.write('<h2 id="tables"> Evaluation Results Table</h2>\n')
        whole_table = _get_table(files=files, file_table=f"table_{lpo_lco_ldo}.html")
        _write_table(f=f, table=whole_table, prefix=prefix)

        if lpo_lco_ldo != "LCO":
            f.write("<h2> Evaluation Results per Cell Line Table</h2>\n")
            cell_line_table = _get_table(files=files, file_table=f"table_cell_line_{lpo_lco_ldo}.html")
            _write_table(f=f, table=cell_line_table, prefix=prefix)
        if lpo_lco_ldo != "LDO":
            f.write("<h2> Evaluation Results per Drug Table</h2>\n")
            drug_table = _get_table(files=files, file_table=f"table_drug_{lpo_lco_ldo}.html")
            _write_table(f=f, table=drug_table, prefix=prefix)
        return f


def _write_table(f: TextIOWrapper, table: str, prefix: str = ""):
    with open(os.path.join(prefix, table)) as eval_f:
        eval_results = eval_f.readlines()
        eval_results[0] = eval_results[0].replace(
            '<table border="1" class="dataframe">',
            '<table class="display customDataTable" style="width:100%">',
        )
        for line in eval_results:
            f.write(line)


def _get_table(files: list, file_table: str) -> str:
    return [f for f in files if f == file_table][0]



"""Contains the baseline MultiOmicsNeuralNetwork model."""

import warnings

import numpy as np
from sklearn.decomposition import PCA

from drevalpy.datasets.dataset import DrugResponseDataset, FeatureDataset

from ..drp_model import DRPModel
from ..utils import get_multiomics_feature_dataset, load_drug_fingerprint_features
from .utils import FeedForwardNetwork


class MultiOmicsNeuralNetwork(DRPModel):
    """Simple Feedforward Neural Network model with dropout using multiple omics data."""

    cell_line_views = [
        "gene_expression",
        "methylation",
        "mutations",
        "copy_number_variation_gistic",
    ]
    drug_views = ["fingerprints"]
    early_stopping = True

    def __init__(self):
        """
        Initalization method for MultiOmicsNeuralNetwork Model.

        The model and the PCA are initialized to None because they are built later in the build_model method.
        """
        super().__init__()
        self.model = None
        self.hyperparameters = None
        self.pca = None

    @classmethod
    def get_model_name(cls) -> str:
        """
        Returns the model name.

        :returns: MultiOmicsNeuralNetwork
        """
        return "MultiOmicsNeuralNetwork"

    def build_model(self, hyperparameters: dict):
        """
        Builds the model from hyperparameters.

        The model is a simple feedforward neural network with dropout. The PCA is used to reduce the dimensionality of
        the methylation data.

        :param hyperparameters: dictionary containing the hyperparameters units_per_layer, dropout_prob, and
            methylation_pca_components.
        """
        self.hyperparameters = hyperparameters
        self.pca = PCA(n_components=hyperparameters["methylation_pca_components"])

    def train(
        self,
        output: DrugResponseDataset,
        cell_line_input: FeatureDataset,
        drug_input: FeatureDataset | None = None,
        output_earlystopping: DrugResponseDataset | None = None,
        model_checkpoint_dir: str = "",
    ):
        """
        Fits the PCA and trains the model.

        :param output: training data associated with the response output
        :param cell_line_input: cell line omics features
        :param drug_input: drug omics features
        :param output_earlystopping: optional early stopping dataset
        :param model_checkpoint_dir: directory to save the model checkpoints
        :raises ValueError: if drug_input (fingerprints) is missing
        """
        if drug_input is None:
            raise ValueError("Drug input (fingerprints) is needed for the MultiOmicsNeuralNetwork model.")

        unique_methylation = np.stack(
            [cell_line_input.features[id_]["methylation"] for id_ in np.unique(output.cell_line_ids)],
            axis=0,
        )

        self.pca.n_components = min(self.pca.n_components, len(unique_methylation))
        self.pca = self.pca.fit(unique_methylation)

        first_feature = next(iter(cell_line_input.features.values()))
        dim_gex = first_feature["gene_expression"].shape[0]
        dim_met = self.pca.n_components
        dim_mut = first_feature["mutations"].shape[0]
        dim_cnv = first_feature["copy_number_variation_gistic"].shape[0]
        dim_fingerprint = next(iter(drug_input.features.values()))["fingerprints"].shape[0]

        self.model = FeedForwardNetwork(
            hyperparameters=self.hyperparameters,
            input_dim=dim_gex + dim_met + dim_mut + dim_cnv + dim_fingerprint,
        )

        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message=".*does not have many workers which may be a bottleneck.*",
            )
            self.model.fit(
                output_train=output,
                cell_line_input=cell_line_input,
                drug_input=drug_input,
                cell_line_views=self.cell_line_views,
                drug_views=self.drug_views,
                output_earlystopping=output_earlystopping,
                batch_size=16,
                patience=5,
                num_workers=1,
                met_transform=self.pca,
                model_checkpoint_dir=model_checkpoint_dir,
            )

    def predict(
        self,
        cell_line_ids: np.ndarray,
        drug_ids: np.ndarray,
        cell_line_input: FeatureDataset,
        drug_input: FeatureDataset | None = None,
    ) -> np.ndarray:
        """
        Transforms the methylation data using the fitted PCA and then predicts the response for the given input.

        :param drug_ids: drug identifiers
        :param cell_line_ids: cell line identifiers
        :param drug_input: drug omics features
        :param cell_line_input: cell line omics features
        :returns: predicted response
        """
        inputs = self.get_feature_matrices(
            cell_line_ids=cell_line_ids,
            drug_ids=drug_ids,
            cell_line_input=cell_line_input,
            drug_input=drug_input,
        )
        (
            gene_expression,
            methylation,
            mutations,
            copy_number_variation_gistic,
            fingerprints,
        ) = (
            inputs["gene_expression"],
            inputs["methylation"],
            inputs["mutations"],
            inputs["copy_number_variation_gistic"],
            inputs["fingerprints"],
        )
        methylation = self.pca.transform(methylation)
        x = np.concatenate(
            (
                gene_expression,
                methylation,
                mutations,
                copy_number_variation_gistic,
                fingerprints,
            ),
            axis=1,
        )
        return self.model.predict(x)

    def load_cell_line_features(self, data_path: str, dataset_name: str) -> FeatureDataset:
        """
        Loads the cell line features.

        :param data_path: data path e.g. data/
        :param dataset_name: dataset name e.g. GDSC1
        :return: FeatureDataset containing the cell line omics features, filtered through the
            drug target genes
        """
        return get_multiomics_feature_dataset(data_path=data_path, dataset_name=dataset_name)

    def load_drug_features(self, data_path: str, dataset_name: str) -> FeatureDataset:
        """
        Load the drug features.

        :param data_path: path to the drug features, in this case the drug fingerprints, e.g., data/
        :param dataset_name: name of the dataset, e.g., GDSC1
        :returns: FeatureDataset containing the drug fingerprint features
        """
        return load_drug_fingerprint_features(data_path, dataset_name)



"""
DIPK model. Adapted from https://github.com/user15632/DIPK.

Original publication:
Improving drug response prediction via integrating gene relationships with deep learning
Pengyong Li, Zhengxiang Jiang, Tianxiao Liu, Xinyu Liu, Hui Qiao, Xiaojun Yao
Briefings in Bioinformatics, Volume 25, Issue 3, May 2024, bbae153, https://doi.org/10.1093/bib/bbae153
"""

import os
import secrets
from typing import Any

import numpy as np
import pandas as pd
import torch
import torch.optim as optim
from torch import nn
from torch.utils.data import DataLoader

from drevalpy.datasets.dataset import DrugResponseDataset, FeatureDataset
from drevalpy.models.drp_model import DRPModel
from drevalpy.models.utils import load_and_reduce_gene_features

from .data_utils import CollateFn, DIPKDataset, get_data, load_bionic_features
from .gene_expression_encoder import GeneExpressionEncoder, encode_gene_expression, train_gene_expession_autoencoder
from .model_utils import Predictor


class DIPKModel(DRPModel):
    """DIPK model. Adapted from https://github.com/user15632/DIPK."""

    cell_line_views = ["gene_expression", "bionic_features"]
    drug_views = ["molgnet_features"]
    early_stopping = True

    def __init__(self) -> None:
        """Initialize the DIPK model."""
        super().__init__()
        self.DEVICE = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        # all of this gets initialized in build_model
        self.model: Predictor | None = None
        self.epochs: int = 0
        self.batch_size: int = 0
        self.lr: float = 0.0
        self.gene_expression_encoder: GeneExpressionEncoder | None = None
        self.epochs_autoencoder: int = 100

    @classmethod
    def get_model_name(cls) -> str:
        """
        Get the model name.

        :returns: DIPK
        """
        return "DIPK"

    def build_model(self, hyperparameters: dict[str, Any]) -> None:
        """
        Builds the DIPK model with the specified hyperparameters.

        :param hyperparameters: embedding_dim, heads, fc_layer_num, fc_layer_dim, dropout_rate, EPOCHS, batch_size, lr

        Details of hyperparameters:

        - embedding_dim: int, embedding dimension used for the graph encoder which is not used in the final model
        - heads: int, number of heads for the multi-head attention layer, defaults to 1
        - fc_layer_num: int, number of fully connected layers for the dense layers
        - fc_layer_dim: list[int], number of neurons for each fully connected layer
        - dropout_rate: float, dropout rate for all fully connected layers
        - EPOCHS: int, number of epochs to train the model
        - batch_size: int, batch size for training
        - lr: float, learning rate for training
        """
        self.model = Predictor(
            hyperparameters["heads"],
            hyperparameters["fc_layer_num"],
            hyperparameters["fc_layer_dim"],
            hyperparameters["dropout_rate"],
        ).to(self.DEVICE)
        self.epochs = hyperparameters["epochs"]
        self.batch_size = hyperparameters["batch_size"]
        self.lr = hyperparameters["lr"]
        self.epochs_autoencoder = hyperparameters["epochs_autoencoder"]
        self.patience = hyperparameters["patience"]

    def train(
        self,
        output: DrugResponseDataset,
        cell_line_input: FeatureDataset,
        drug_input: FeatureDataset | None = None,
        output_earlystopping: DrugResponseDataset | None = None,
        model_checkpoint_dir: str = "checkpoints",
    ) -> None:
        """
        Trains the model.

        :param output: training data associated with the response output
        :param cell_line_input: input data associated with the cell line
        :param drug_input: input data associated with the drug
        :param output_earlystopping: early stopping data associated with the response output
        :param model_checkpoint_dir: directory to save the model checkpoint
        :raises ValueError: if drug_input is None or if the model is not initialized
        """
        if drug_input is None:
            raise ValueError("DIPK model requires drug features.")
        if not isinstance(self.model, Predictor):
            raise ValueError("DIPK model not initialized.")
        if output_earlystopping is None:
            raise ValueError("DIPK model requires early stopping data.")

        loss_func = nn.MSELoss()
        params = [{"params": self.model.parameters()}]
        optimizer = optim.Adam(params, lr=self.lr)

        self.gene_expression_encoder = train_gene_expession_autoencoder(
            cell_line_input.get_feature_matrix(view="gene_expression", identifiers=output.cell_line_ids),
            cell_line_input.get_feature_matrix(view="gene_expression", identifiers=output_earlystopping.cell_line_ids),
            epochs_autoencoder=self.epochs_autoencoder,
        )

        cell_line_input.apply(
            lambda x: encode_gene_expression(x, self.gene_expression_encoder),  # type: ignore[arg-type]
            view="gene_expression",
        )  # type: ignore[arg-type]

        # Load data
        collate = CollateFn(train=True)
        train_samples = get_data(
            cell_ids=output.cell_line_ids,
            drug_ids=output.drug_ids,
            cell_line_features=cell_line_input,
            drug_features=drug_input,
            ic50=output.response,
        )
        early_stopping_samples = get_data(
            cell_ids=output_earlystopping.cell_line_ids,
            drug_ids=output_earlystopping.drug_ids,
            cell_line_features=cell_line_input,
            drug_features=drug_input,
            ic50=output_earlystopping.response,
        )

        train_loader: DataLoader = DataLoader(
            DIPKDataset(train_samples), batch_size=self.batch_size, shuffle=True, collate_fn=collate
        )
        early_stopping_loader: DataLoader = DataLoader(
            DIPKDataset(early_stopping_samples), batch_size=self.batch_size, shuffle=True, collate_fn=collate
        )

        # Early stopping parameters
        best_val_loss = float("inf")
        epochs_without_improvement = 0

        # Ensure the checkpoint directory exists
        os.makedirs(model_checkpoint_dir, exist_ok=True)
        version = "version-" + "".join(
            [secrets.choice("0123456789abcdef") for _ in range(20)]
        )  # preventing conflicts of filenames

        checkpoint_path = os.path.join(model_checkpoint_dir, f"{version}_best_DIPK_model.pth")

        # Train model
        print("Training DIPK model")
        for epoch in range(self.epochs):
            self.model.train()
            epoch_loss = 0.0
            batch_count = 0

            # Training phase
            for batch in train_loader:
                drug_features = batch["molgnet_features"].to(self.DEVICE)
                gene_features = batch["gene_features"].to(self.DEVICE)
                bionic_features = batch["bionic_features"].to(self.DEVICE)
                molgnet_mask = batch["molgnet_mask"].to(self.DEVICE)
                ic50_values = batch["ic50_values"].to(self.DEVICE)

                # Forward pass
                prediction = self.model(
                    molgnet_drug_features=drug_features,
                    gene_expression=gene_features,
                    bionic=bionic_features,
                    molgnet_mask=molgnet_mask,
                )

                # Compute the loss
                loss = loss_func(torch.squeeze(prediction), torch.squeeze(ic50_values))

                # Backpropagation
                optimizer.zero_grad()
                loss.backward()
                optimizer.step()

                # Update loss and batch count
                epoch_loss += loss.detach().item()
                batch_count += 1

            epoch_loss /= batch_count
            print(f"DIPK: Epoch [{epoch + 1}] Training Loss: {epoch_loss:.4f}")

            # Validation phase for early stopping
            self.model.eval()
            val_loss = 0.0
            val_batch_count = 0
            with torch.no_grad():
                for batch in early_stopping_loader:
                    drug_features = batch["molgnet_features"].to(self.DEVICE)
                    gene_features = batch["gene_features"].to(self.DEVICE)
                    bionic_features = batch["bionic_features"].to(self.DEVICE)
                    molgnet_mask = batch["molgnet_mask"].to(self.DEVICE)
                    ic50_values = batch["ic50_values"].to(self.DEVICE)

                    # Forward pass
                    prediction = self.model(
                        molgnet_drug_features=drug_features,
                        gene_expression=gene_features,
                        bionic=bionic_features,
                        molgnet_mask=molgnet_mask,
                    )

                    # Compute the loss
                    loss = loss_func(torch.squeeze(prediction), torch.squeeze(ic50_values))

                    # Update validation loss
                    val_loss += loss.item()
                    val_batch_count += 1

            val_loss /= val_batch_count
            print(f"DIPK: Epoch [{epoch + 1}] Validation Loss: {val_loss:.4f}")

            # Checkpointing: Save the best model
            if val_loss < best_val_loss:
                best_val_loss = val_loss
                epochs_without_improvement = 0
                # Save the model checkpoint securely
                torch.save(self.model.state_dict(), checkpoint_path)  # noqa S614
                print(f"DIPK: Saved best model at epoch {epoch + 1}")
            else:
                epochs_without_improvement += 1
                if epochs_without_improvement >= self.patience:
                    print(f"DIPK: Early stopping triggered at epoch {epoch + 1}")
                    break

        # Reload the best model after training
        print("DIPK: Reloading the best model")
        self.model.load_state_dict(
            torch.load(checkpoint_path, map_location=self.DEVICE, weights_only=True)  # noqa S614
        )
        self.model.to(self.DEVICE)  # Ensure model is on the correct device

    def predict(
        self,
        cell_line_ids: np.ndarray,
        drug_ids: np.ndarray,
        cell_line_input: FeatureDataset,
        drug_input: FeatureDataset | None = None,
    ) -> np.ndarray:
        """
        Predicts the response values for the given cell lines and drugs.

        :param cell_line_ids: list of cell line IDs
        :param drug_ids: list of drug IDs
        :param cell_line_input: input data associated with the cell line
        :param drug_input: input data associated with the drug
        :return: predicted response values
        :raises ValueError: if drug_input is None or if the model is not initialized
        """
        if drug_input is None:
            raise ValueError("DIPK model requires drug features.")
        if not isinstance(self.model, Predictor):
            raise ValueError("DIPK model not initialized.")

        # Load data
        collate = CollateFn(train=False)
        test_samples = get_data(
            cell_ids=cell_line_ids,
            drug_ids=drug_ids,
            cell_line_features=cell_line_input,
            drug_features=drug_input,
        )
        test_loader: DataLoader = DataLoader(
            DIPKDataset(test_samples), batch_size=self.batch_size, shuffle=False, collate_fn=collate
        )

        # Run prediction
        self.model.eval()
        predictions = []
        with torch.no_grad():
            for batch in test_loader:
                drug_features = batch["molgnet_features"].to(self.DEVICE)
                gene_features = batch["gene_features"].to(self.DEVICE)
                bionic_features = batch["bionic_features"].to(self.DEVICE)
                molgnet_mask = batch["molgnet_mask"].to(self.DEVICE)

                prediction = self.model(
                    molgnet_drug_features=drug_features,
                    gene_expression=gene_features,
                    bionic=bionic_features,
                    molgnet_mask=molgnet_mask,
                )
                predictions += torch.squeeze(prediction).cpu().tolist()

        return np.array(predictions)

    def load_cell_line_features(self, data_path: str, dataset_name: str) -> FeatureDataset:
        """
        Load cell line features.

        :param data_path: path to the data
        :param dataset_name: path to the dataset
        :returns: cell line features
        """
        gene_expression = load_and_reduce_gene_features(
            feature_type="gene_expression",
            gene_list=None,
            data_path=data_path,
            dataset_name=dataset_name,
        )
        bionic_features = load_bionic_features(
            data_path=data_path,
            dataset_name=dataset_name,
        )
        bionic_features.add_features(gene_expression)

        return bionic_features

    def load_drug_features(self, data_path: str, dataset_name: str) -> FeatureDataset:
        """
        Load drug features.

        :param data_path: path to the data
        :param dataset_name: path to the dataset
        :returns: drug features
        """

        def load_feature(file_path, sep="\t"):
            return np.array(pd.read_csv(file_path, index_col=0, sep=sep))

        drug_path = os.path.join(data_path, dataset_name, "DIPK_features", "Drugs")
        files_in_drug_path = os.listdir(drug_path)
        drug_list = [
            file.split("_")[1].split(".csv")[0]
            for file in files_in_drug_path
            if file.endswith(".csv") and file.startswith("MolGNet")
        ]

        return FeatureDataset(
            features={
                drug: {
                    "molgnet_features": load_feature(os.path.join(drug_path, f"MolGNet_{drug}.csv")),
                }
                for drug in drug_list
            }
        )



"""Parent class for Violin and Heatmap plots of performance measures over CV runs."""

from io import TextIOWrapper

import pandas as pd

from drevalpy.visualization.outplot import OutPlot


class VioHeat(OutPlot):
    """Parent class for Violin and Heatmap plots of performance measures over CV runs."""

    def __init__(self, df: pd.DataFrame, normalized_metrics=False, whole_name=False):
        """
        Initialize the VioHeat class.

        :param df: evaluation results, either overall or per algorithm
        :param normalized_metrics: whether the metrics are normalized
        :param whole_name: whether the whole name should be displayed
        """
        self.df = df.sort_index()
        self.all_metrics = [
            "R^2",
            "R^2: drug normalized",
            "R^2: cell_line normalized",
            "Pearson",
            "Pearson: drug normalized",
            "Pearson: cell_line normalized",
            "Spearman",
            "Spearman: drug normalized",
            "Spearman: cell_line normalized",
            "Kendall",
            "Kendall: drug normalized",
            "Kendall: cell_line normalized",
            "Partial_Correlation",
            "Partial_Correlation: drug normalized",
            "Partial_Correlation: cell_line normalized",
            "MSE",
            "RMSE",
            "MAE",
        ]
        self.normalized_metrics = normalized_metrics
        self.whole_name = whole_name
        if self.normalized_metrics:
            self.all_metrics = [metric for metric in self.all_metrics if "normalized" in metric]
        else:
            self.all_metrics = [metric for metric in self.all_metrics if "normalized" not in metric]

    def draw_and_save(self, out_prefix: str, out_suffix: str) -> None:
        """
        Draw and save the plot.

        :param out_prefix: e.g., results/my_run/heatmaps/
        :param out_suffix: e.g., algorithms_normalized
        """
        pass

    def _draw(self) -> None:
        pass

    @staticmethod
    def write_to_html(lpo_lco_ldo: str, f: TextIOWrapper, *args, **kwargs) -> TextIOWrapper:
        """
        Write the Violin and Heatmap plots into the result HTML file.

        :param lpo_lco_ldo: setting, e.g., LPO
        :param f: result HTML file
        :param args: additional arguments
        :param kwargs: additional keyword arguments, in this case, the plot type and the files
        :returns: the result HTML file
        """
        plot: str = kwargs.get("plot", "")
        files: list[str] = kwargs.get("files", [])

        if plot == "Violin":
            nav_id = "violin"
            dir_name = "violin_plots"
            prefix = "violin"
        else:
            nav_id = "heatmap"
            dir_name = "heatmaps"
            prefix = "heatmap"
        plot_list = [
            f
            for f in files
            if (
                lpo_lco_ldo in f
                and f.startswith(prefix)
                and f != f"{prefix}_{lpo_lco_ldo}.html"
                and f != f"{prefix}_{lpo_lco_ldo}_normalized.html"
            )
        ]
        f.write(f"<h2 id={nav_id!r}>{plot} Plots of Performance Measures over CV runs</h2>\n")
        f.write(f"<h3>{plot} plots comparing all models</h3>\n")
        f.write(
            f'<iframe src="{dir_name}/{prefix}_algorithms_{lpo_lco_ldo}.html" width="100%" height="100%" '
            f'frameBorder="0"></iframe>\n'
        )
        f.write(f"<h3>{plot} plots comparing all models with normalized metrics</h3>\n")
        f.write(
            "Before calculating the evaluation metrics, all values were normalized by the mean of the drug or cell "
            "line. Since this only influences the R^2 and the correlation metrics, the error metrics are not shown. \n"
        )
        f.write(
            f'<iframe src="{dir_name}/{prefix}_algorithms_{lpo_lco_ldo}_normalized.html" width="100%" height="100%" '
            f'frameBorder="0"></iframe>\n'
        )
        f.write(f"<h3>{plot} plots comparing performance measures for tests within each model</h3>\n")
        f.write("<ul>")
        for plot in plot_list:
            f.write(f'<li><a href="{dir_name}/{plot}" target="_blank">{plot}</a></li>\n')
        f.write("</ul>\n")
        return f
